/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/AIMath.h"
#include <stdbool.h>

static bool intersectLineSegments(Vector2 point0, Vector2 point1, Vector2 point2, Vector2 point3) {
	return (Vector2_cross(Vector2_subtract(point1, point0), Vector2_subtract(point2, point0)) > 0.0f) !=
	       (Vector2_cross(Vector2_subtract(point1, point0), Vector2_subtract(point3, point0)) > 0.0f) &&
	       (Vector2_cross(Vector2_subtract(point3, point2), Vector2_subtract(point0, point2)) > 0.0f) !=
	       (Vector2_cross(Vector2_subtract(point3, point2), Vector2_subtract(point1, point2)) > 0.0f);
}

static bool intersectLineRect(Vector2 endpoint0, Vector2 endpoint1, float left, float right, float bottom, float top) {
	return intersectLineSegments(endpoint0, endpoint1, (Vector2) {left, bottom}, (Vector2) {right, bottom}) ||
	       intersectLineSegments(endpoint0, endpoint1, (Vector2) {left, bottom}, (Vector2) {left, top}) ||
	       intersectLineSegments(endpoint0, endpoint1, (Vector2) {left, top}, (Vector2) {right, top}) ||
	       intersectLineSegments(endpoint0, endpoint1, (Vector2) {right, bottom}, (Vector2) {right, top});
}

static bool occludesLOS(GameObjectType type) {
	switch (type) {
		case GAME_OBJECT_TYPE_DOOR:
			return true;
		default:
			return false;
	}
}

float indexDistance(int index1, int index2, size_t rowWidth) {
	return (index1 % rowWidth - index2 % rowWidth) * (index1 % rowWidth - index2 % rowWidth) + (index1 / rowWidth - index2 / rowWidth) * (index1 / rowWidth - index2 / rowWidth);
}

GameObjectModel ** AIObjectsInLineOfSight(Vector2 referencePoint,
                                          Vector2 direction,
                                          float fov,
                                          float distance,
                                          struct DungeonTile * tiles,
                                          size_t tileCountX,
                                          size_t tileCountY,
                                          GameObjectModel ** gameObjects,
                                          size_t gameObjectCount,
                                          size_t * outObjectCount) {
	size_t objectIndex, objectIndex2;
	struct rect4f objectRect, objectRect2, tileRect;
	Vector2 center;
	GameObjectModel ** objects;
	size_t objectCount;
	Vector2 objectDirection;
	float objectDistance;
	bool outOfLOS;
	
	objectCount = 0;
	objects = malloc(sizeof(GameObjectModel *) * gameObjectCount);
	for (objectIndex = 0; objectIndex < gameObjectCount; objectIndex++) {
		objectRect = gameObjects[objectIndex]->roughBounds(gameObjects[objectIndex]);
		center.x = objectRect.left + (objectRect.right - objectRect.left) * 0.5f;
		center.y = objectRect.bottom + (objectRect.top - objectRect.bottom) * 0.5f;
		objectDirection = Vector2_subtract(center, referencePoint);
		objectDistance = (center.x - referencePoint.x) * (center.x - referencePoint.x) + (center.y - referencePoint.y) * (center.y - referencePoint.y);
		Vector2_normalize(&objectDirection);
		if (objectDistance < distance * distance && acos(Vector2_dot(objectDirection, direction)) < fov * 0.5f) {
			unsigned int tileIndexX, tileIndexY;
			
			tileRect.left = floor(min(center.x, referencePoint.x));
			tileRect.right = ceil(max(center.x, referencePoint.x));
			tileRect.bottom = floor(min(center.y, referencePoint.y));
			tileRect.top = ceil(max(center.y, referencePoint.y));
			
			if (tileRect.left < 0) {
				tileRect.left = 0;
			} else if (tileRect.left > tileCountX) {
				tileRect.left = tileCountX;
			}
			if (tileRect.right < 0) {
				tileRect.right = 0;
			} else if (tileRect.right > tileCountX) {
				tileRect.right = tileCountX;
			}
			if (tileRect.bottom < 0) {
				tileRect.bottom = 0;
			} else if (tileRect.left > tileCountY) {
				tileRect.bottom = tileCountY;
			}
			if (tileRect.top < 0) {
				tileRect.top = 0;
			} else if (tileRect.top > tileCountY) {
				tileRect.top = tileCountY;
			}
			
			outOfLOS = false;
			for (tileIndexY = (unsigned int) tileRect.bottom; tileIndexY < (unsigned int) tileRect.top && !outOfLOS; tileIndexY++) {
				for (tileIndexX = (unsigned int) tileRect.left; tileIndexX < (unsigned int) tileRect.right && !outOfLOS; tileIndexX++) {
					if (tiles[tileIndexY * tileCountX + tileIndexX].type == DUNGEON_TILE_TYPE_WALL &&
					    intersectLineRect(center, referencePoint, tileIndexX, tileIndexX + 1.0f, tileIndexY, tileIndexY + 1.0f)) {
						outOfLOS = true;
					}
				}
			}
			if (!outOfLOS) {
				objects[objectCount++] = gameObjects[objectIndex];
			}
		}
	}
	
	for (objectIndex = 0; objectIndex < objectCount; objectIndex++) {
		objectRect = objects[objectIndex]->roughBounds(objects[objectIndex]);
		center.x = objectRect.left + (objectRect.right - objectRect.left) * 0.5f;
		center.y = objectRect.bottom + (objectRect.top - objectRect.bottom) * 0.5f;
		outOfLOS = false;
		
		for (objectIndex2 = 0; objectIndex2 < objectCount; objectIndex2++) {
			if (objectIndex2 == objectIndex) {
				continue;
			}
			if (occludesLOS(objects[objectIndex2]->type)) {
				objectRect2 = objects[objectIndex2]->roughBounds(objects[objectIndex2]);
				if (intersectLineRect(center, referencePoint, objectRect2.left, objectRect2.right, objectRect2.bottom, objectRect2.top)) {
					outOfLOS = true;
					break;
				}
			}
		}
		
		if (outOfLOS) {
			objectCount--;
			for (objectIndex2 = objectIndex; objectIndex2 < objectCount; objectIndex2++) {
				objects[objectIndex2] = objects[objectIndex2 + 1];
			}
			objectIndex--;
		}
	}
	
	*outObjectCount = objectCount;
	return objects;
}

struct travelNode {
	int index;
	int previousIndex;
	int steps;
	float destinationDistance;
};

struct travelPath AIBestPathToDestination(int origin,
                                          int destination,
                                          struct DungeonTile * tiles,
                                          size_t tileCountX,
                                          size_t tileCountY) {
	struct travelPath path;
	struct travelNode * nodeQueue, * visitedNodes;
	size_t nodeQueueCount, visitedNodeCount, nodeIndex, nodeIndex2, neighborIndex;
	const int neighborOffsets[8] = {
		-1 - tileCountX, -tileCountX, 1 - tileCountX,
		-1,                           1,
		-1 + tileCountX,  tileCountX, 1 + tileCountX
	};
	struct travelNode currentNode, neighborNode;
	int tileIndexX, tileIndexY;
	bool wall, visited, queued;
	
	if (destination == origin) {
		path.nodeCount = 0;
		path.nodes = NULL;
		return path;
	}
	
	nodeQueue = malloc(sizeof(struct travelNode) * tileCountX * tileCountY);
	nodeQueueCount = 1;
	nodeQueue[0].index = origin;
	nodeQueue[0].previousIndex = -1;
	nodeQueue[0].steps = 0;
	nodeQueue[0].destinationDistance = indexDistance(origin, destination, tileCountX);
	visitedNodes = malloc(sizeof(struct travelNode) * tileCountX * tileCountY);
	visitedNodeCount = 0;
	
	while (nodeQueueCount > 0) {
		currentNode = nodeQueue[0];
		if (currentNode.index == destination) {
			int * reverseNodeIndexMap;
			
			reverseNodeIndexMap = malloc(sizeof(int) * tileCountX * tileCountY);
			for (nodeIndex = 0; nodeIndex < visitedNodeCount; nodeIndex++) {
				reverseNodeIndexMap[visitedNodes[nodeIndex].index] = nodeIndex;
			}
			
			path.nodeCount = currentNode.steps;
			path.nodes = malloc(sizeof(int) * path.nodeCount);
			do {
				path.nodes[currentNode.steps - 1] = currentNode.index;
				currentNode = visitedNodes[reverseNodeIndexMap[currentNode.previousIndex]];
			} while (currentNode.previousIndex != -1);
			
			free(reverseNodeIndexMap);
			free(nodeQueue);
			free(visitedNodes);
			return path;
			
		} else {
			visitedNodes[visitedNodeCount++] = currentNode;
			nodeQueueCount--;
			for (nodeIndex = 0; nodeIndex < nodeQueueCount; nodeIndex++) {
				nodeQueue[nodeIndex] = nodeQueue[nodeIndex + 1];
			}
			
			for (neighborIndex = 0; neighborIndex < 8; neighborIndex++) {
				neighborNode.index = currentNode.index + neighborOffsets[neighborIndex];
				if ((currentNode.index % tileCountX == 0 && (neighborIndex == 0 || neighborIndex == 3 || neighborIndex == 5)) ||
				    (currentNode.index % tileCountX == tileCountX - 1 && (neighborIndex == 2 || neighborIndex == 4 || neighborIndex == 7)) ||
				    neighborNode.index < 0 ||
				    neighborNode.index >= (int) (tileCountX * tileCountY)) {
					continue;
				}
				wall = false;
				for (tileIndexY = min(neighborNode.index / tileCountX, currentNode.index / tileCountX); tileIndexY <= (int) (max(neighborNode.index / tileCountX, currentNode.index / tileCountX)); tileIndexY++) {
					for (tileIndexX = min(neighborNode.index % tileCountX, currentNode.index % tileCountX); tileIndexX <= (int) (max(neighborNode.index % tileCountX, currentNode.index % tileCountX)); tileIndexX++) {
						if (tiles[tileIndexY * tileCountX + tileIndexX].type == DUNGEON_TILE_TYPE_WALL) {
							wall = true;
							break;
						}
					}
					if (wall) {break;}
				}
				if (wall) {continue;}
				
				visited = false;
				for (nodeIndex = 0; nodeIndex < visitedNodeCount; nodeIndex++) {
					if (visitedNodes[nodeIndex].index == neighborNode.index) {
						visited = true;
						break;
					}
				}
				if (visited) {continue;}
				
				queued = false;
				for (nodeIndex = 0; nodeIndex < nodeQueueCount; nodeIndex++) {
					if (nodeQueue[nodeIndex].index == neighborNode.index) {
						queued = true;
						break;
					}
				}
				if (queued) {continue;}
				
				neighborNode.previousIndex = currentNode.index;
				neighborNode.steps = currentNode.steps + 1;
				neighborNode.destinationDistance = indexDistance(neighborNode.index, destination, tileCountX);
				for (nodeIndex = 0; nodeIndex < nodeQueueCount; nodeIndex++) {
					if (neighborNode.destinationDistance < nodeQueue[nodeIndex].destinationDistance) {
						for (nodeIndex2 = nodeQueueCount; nodeIndex2 > nodeIndex; nodeIndex2--) {
							nodeQueue[nodeIndex2] = nodeQueue[nodeIndex2 - 1];
						}
						nodeQueueCount++;
						nodeQueue[nodeIndex] = neighborNode;
						break;
					}
				}
				if (nodeIndex == nodeQueueCount) {
					nodeQueue[nodeQueueCount++] = neighborNode;
				}
			}
		}
	}
	
	free(nodeQueue);
	free(visitedNodes);
	path.nodeCount = 0;
	path.nodes = NULL;
	return path;
}
