/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/DungeonRoomView.h"
#include "glgraphics/GLGraphics.h"
#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

struct vertex_p2f_t2f {
	GLfloat position[2];
	GLfloat texCoords[2];
};

DungeonRoomView * DungeonRoomView_create(compat_type(DungeonRoomStateModel *) stateModel, ResourceManager * resourceManager) {
	stemobject_create_implementation(DungeonRoomView, init, stateModel, resourceManager);
}

#define TEXEL_INSET 0.125f

void DungeonRoomView_init(compat_type(DungeonRoomView *) selfPtr, compat_type(DungeonRoomStateModel *) stateModel, ResourceManager * resourceManager) {
	DungeonRoomView * self = selfPtr;
	struct vertex_p2f_t2f * vertices;
	GLushort * indexes;
	unsigned int tileIndex;
	size_t tileCount, canvasTileIndex;
	float positionX, positionY;
	float texLeft = 0.0f, texRight = 0.0f, texBottom = 0.0f, texTop = 0.0f;
	
	StemObject_init(self);
	self->dispose = DungeonRoomView_dispose;
	self->draw = DungeonRoomView_draw;
	
	self->resourceManager = resourceManager;
	self->stateModel = stateModel;
	glGenBuffers(1, &self->vboID);
	glGenBuffers(1, &self->eboID);
	
	self->tileTextureCanvas = self->resourceManager->referenceResource(self->resourceManager, "texture_canvas", self->stateModel->tileCanvasFileName);
	self->tileTexture = self->resourceManager->referenceResource(self->resourceManager, "texture", self->tileTextureCanvas->textureFileName);
	
	tileCount = self->stateModel->tileCountX * self->stateModel->tileCountY;
	self->numVertices = tileCount * 4;
	self->numIndexes = tileCount * 6;
	vertices = malloc(sizeof(struct vertex_p2f_t2f) * self->numVertices);
	indexes = malloc(sizeof(GLushort) * self->numIndexes);
	for (tileIndex = 0; tileIndex < tileCount; tileIndex++) {
		for (canvasTileIndex = 0; canvasTileIndex < self->tileTextureCanvas->tileCount; canvasTileIndex++) {
			if (self->tileTextureCanvas->tiles[canvasTileIndex].name == self->stateModel->tiles[tileIndex].canvasTileID) {
				texLeft = (self->tileTextureCanvas->tiles[canvasTileIndex].bounds.left + TEXEL_INSET) / self->tileTexture->pixelWidth;
				texRight = (self->tileTextureCanvas->tiles[canvasTileIndex].bounds.right - TEXEL_INSET) / self->tileTexture->pixelWidth;
				texBottom = (self->tileTextureCanvas->tiles[canvasTileIndex].bounds.bottom + TEXEL_INSET) / self->tileTexture->pixelHeight;
				texTop = (self->tileTextureCanvas->tiles[canvasTileIndex].bounds.top - TEXEL_INSET) / self->tileTexture->pixelHeight;
				break;
			}
		}
		if (canvasTileIndex == self->tileTextureCanvas->tileCount) {
			fprintf(stderr, "Warning: This room uses canvas tile \"%s\", which is not defined in its corresponding tile canvas file\n", self->stateModel->tiles[tileIndex].canvasTileID);
			texLeft = texBottom = 0.0f;
			texRight = texTop = 1.0f;
		}
		positionX = tileIndex % self->stateModel->tileCountX;
		positionY = tileIndex / self->stateModel->tileCountX;
		
		vertices[tileIndex * 4 + 0].position[0] = positionX;
		vertices[tileIndex * 4 + 0].position[1] = positionY;
		vertices[tileIndex * 4 + 0].texCoords[0] = texLeft;
		vertices[tileIndex * 4 + 0].texCoords[1] = texBottom;
		vertices[tileIndex * 4 + 1].position[0] = positionX + TILE_SIZE;
		vertices[tileIndex * 4 + 1].position[1] = positionY;
		vertices[tileIndex * 4 + 1].texCoords[0] = texRight;
		vertices[tileIndex * 4 + 1].texCoords[1] = texBottom;
		vertices[tileIndex * 4 + 2].position[0] = positionX + TILE_SIZE;
		vertices[tileIndex * 4 + 2].position[1] = positionY + TILE_SIZE;
		vertices[tileIndex * 4 + 2].texCoords[0] = texRight;
		vertices[tileIndex * 4 + 2].texCoords[1] = texTop;
		vertices[tileIndex * 4 + 3].position[0] = positionX;
		vertices[tileIndex * 4 + 3].position[1] = positionY + TILE_SIZE;
		vertices[tileIndex * 4 + 3].texCoords[0] = texLeft;
		vertices[tileIndex * 4 + 3].texCoords[1] = texTop;
		
		indexes[tileIndex * 6 + 0] = tileIndex * 4 + 0;
		indexes[tileIndex * 6 + 1] = tileIndex * 4 + 1;
		indexes[tileIndex * 6 + 2] = tileIndex * 4 + 2;
		indexes[tileIndex * 6 + 3] = tileIndex * 4 + 2;
		indexes[tileIndex * 6 + 4] = tileIndex * 4 + 3;
		indexes[tileIndex * 6 + 5] = tileIndex * 4 + 0;
	}
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_t2f) * self->numVertices, vertices, GL_STATIC_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * self->numIndexes, indexes, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	free(vertices);
	free(indexes);
}

void DungeonRoomView_dispose(compat_type(DungeonRoomView *) selfPtr) {
	DungeonRoomView * self = selfPtr;
	
	glDeleteBuffers(1, &self->vboID);
	glDeleteBuffers(1, &self->eboID);
	self->resourceManager->releaseResource(self->resourceManager, "texture", self->tileTextureCanvas->textureFileName);
	self->resourceManager->releaseResource(self->resourceManager, "texture_canvas", self->stateModel->tileCanvasFileName);
	StemObject_dispose(selfPtr);
}

void DungeonRoomView_draw(compat_type(DungeonRoomView *) selfPtr, Matrix modelviewMatrix) {
	DungeonRoomView * self = selfPtr;
	
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	
	glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_t2f), (void *) offsetof(struct vertex_p2f_t2f, position));
	glTexCoordPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_t2f), (void *) offsetof(struct vertex_p2f_t2f, texCoords));
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	
	self->tileTexture->activate(self->tileTexture);
	glDrawElements(GL_TRIANGLES, self->numIndexes, GL_UNSIGNED_SHORT, 0);
	self->tileTexture->deactivate(self->tileTexture);
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
