/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/EnemyBerserkerModel.h"
#include "convergence/SwordSwingModel.h"
#include <stdlib.h>

#define ATTACK_DISTANCE_RANK_0 2.0f
#define ATTACK_DISTANCE_RANK_1 2.125f
#define ATTACK_DISTANCE_RANK_2 2.375f

EnemyBerserkerModel * EnemyBerserkerModel_create(DungeonRoomStateModel * stateModel, int rank, Vector2 position, Direction facing, size_t inventoryItemCount, struct InventoryModel_item * inventoryItems, size_t patrolWaypointCount, Vector2 * patrolWaypoints) {
	stemobject_create_implementation(EnemyBerserkerModel, init, stateModel, rank, position, facing, inventoryItemCount, inventoryItems, patrolWaypointCount, patrolWaypoints)
}

static void sharedInit(EnemyBerserkerModel * self) {
	self->dispose = EnemyBerserkerModel_dispose;
	self->copy = EnemyBerserkerModel_copy;
	self->update = EnemyBerserkerModel_update;
	self->protected_ivar(aiUpdate) = EnemyBerserkerModel_aiUpdate;
	self->atRest = EnemyBerserkerModel_atRest;
	self->setAIState = EnemyBerserkerModel_setAIState;
	self->attack = EnemyBerserkerModel_attack;
	
	self->enemyType = ENEMY_TYPE_BERSERKER;
	self->acceleration = 1.0f / 512.0f;
	self->deceleration = 2.0f / 512.0f;
	self->velocityMax = 12.0f / 512.0f;
}

void EnemyBerserkerModel_init(compat_type(EnemyBerserkerModel *) selfPtr, DungeonRoomStateModel * stateModel, int rank, Vector2 position, Direction facing, size_t inventoryItemCount, struct InventoryModel_item * inventoryItems, size_t patrolWaypointCount, Vector2 * patrolWaypoints) {
	EnemyBerserkerModel * self = selfPtr;
	int health;
	
	switch (rank) {
		case 0:
		default:
			health = 4;
			break;
		case 1:
			health = 14;
			break;
		case 2:
			health = 45;
			break;
	}
	EnemyModel_init(self, stateModel, rank, health, position, facing, inventoryItemCount, inventoryItems, patrolWaypointCount, patrolWaypoints);
	sharedInit(self);
	self->attackCooldown = 0;
	if (rank == 2) {
		self->knockbackResistance = 0.375f;
	}
}

void EnemyBerserkerModel_initCopy(compat_type(EnemyBerserkerModel *) selfPtr, compat_type(EnemyBerserkerModel *) originalPtr) {
	EnemyBerserkerModel * self = selfPtr, * original = originalPtr;
	
	EnemyModel_initCopy(self, original);
	sharedInit(self);
	
	self->attackCooldown = original->attackCooldown;
	self->swordSwingFacing = original->swordSwingFacing;
}

void EnemyBerserkerModel_dispose(compat_type(EnemyBerserkerModel *) selfPtr) {
	EnemyBerserkerModel * self = selfPtr;
	
	EnemyModel_dispose(self);
}

compat_type(EnemyBerserkerModel *) EnemyBerserkerModel_copy(compat_type(EnemyBerserkerModel *) selfPtr) {
	stemobject_copy_implementation(EnemyBerserkerModel, initCopy)
}

void EnemyBerserkerModel_update(compat_type(EnemyBerserkerModel *) selfPtr) {
	EnemyBerserkerModel * self = selfPtr;
	
	EnemyModel_update(self);
	
	if (self->aiState == ENEMY_AI_STATE_ATTACK) {
		self->facing = vectorToNearestDirection(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
		self->attack(self);
	}
	
	if (self->attackCooldown > 0) {
		self->attackCooldown--;
		self->facing = self->swordSwingFacing;
	}
}

void EnemyBerserkerModel_aiUpdate(compat_type(EnemyBerserkerModel *) selfPtr) {
	EnemyBerserkerModel * self = selfPtr;
	
	EnemyModel_aiUpdate(self);
	if (self->canSeePlayer) {
		float attackDistance;
		
		switch (self->rank) {
			case 0:
			default:
				attackDistance = ATTACK_DISTANCE_RANK_0;
				break;
			case 1:
				attackDistance = ATTACK_DISTANCE_RANK_1;
				break;
			case 2:
				attackDistance = ATTACK_DISTANCE_RANK_2;
				break;
		}
		if ((self->position.x - self->lastKnownPlayerPosition.x) * (self->position.x - self->lastKnownPlayerPosition.x) + (self->position.y - self->lastKnownPlayerPosition.y) * (self->position.y - self->lastKnownPlayerPosition.y) < attackDistance * attackDistance) {
			self->setAIState(self, ENEMY_AI_STATE_ATTACK);
		} else {
			self->setAIState(self, ENEMY_AI_STATE_PURSUE);
		}
	}
}

bool EnemyBerserkerModel_atRest(compat_type(EnemyBerserkerModel *) selfPtr) {
	EnemyBerserkerModel * self = selfPtr;
	
	return self->attackCooldown <= 0 &&
	       EnemyModel_atRest(self);
}

void EnemyBerserkerModel_setAIState(compat_type(EnemyBerserkerModel *) selfPtr, enum EnemyAIState state) {
	EnemyBerserkerModel * self = selfPtr;
	
	EnemyModel_setAIState(self, state);
	switch (state) {
		case ENEMY_AI_STATE_PURSUE:
			self->acceleration = 2.0f / 512.0f;
			self->deceleration = 4.0f / 512.0f;
			switch (self->rank) {
				case 0:
				default:
					self->velocityMax = 17.0f / 512.0f;
					break;
				case 1:
					self->velocityMax = 19.0f / 512.0f;
					break;
				case 2:
					self->velocityMax = 21.0f / 512.0f;
					break;
			}
			break;
			
		case ENEMY_AI_STATE_PATROL:
		default:
			self->acceleration = 1.0f / 512.0f;
			self->deceleration = 2.0f / 512.0f;
			self->velocityMax = 12.0f / 512.0f;
			break;
	}
}

void EnemyBerserkerModel_attack(compat_type(EnemyBerserkerModel *) selfPtr) {
	EnemyBerserkerModel * self = selfPtr;
	
	if (self->attackCooldown <= 0 && self->stun <= 0) {
		self->swordSwingFacing = self->facing;
		switch (self->rank) {
			case 0:
			default:
				self->attackCooldown = 60;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, true, 1, 0.09375f, 0.5f, 1.75f, 0.5f, 24));
				break;
			case 1:
				self->attackCooldown = 50;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, true, 3, 0.125f, 0.5f, 2.0f, 0.5f, 24));
				break;
			case 2:
				self->attackCooldown = 35;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, true, 5, 0.15625f, 0.5f, 2.25f, 0.5f, 20));
				break;
		}
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(ENEMY_EVENT_ATTACK), NULL);
	}
}
