/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/EnemyFirstBossModel.h"
#include "convergence/ProjectileModel.h"
#include "convergence/SwordSwingModel.h"
#include <stdlib.h>

#define ATTACK_DISTANCE 10.0f
static const int DAMAGES[3] = {
	3,
	4,
	6
};
static const float SHOT_SPEEDS[3] = {
	0.075f,
	0.0875f,
	0.10625f
};
static const float KNOCKBACKS[3] = {
	0.1875f,
	0.28125f,
	0.375f
};
static const int SHOTS_BEFORE_BREAKS[3] = {
	6,
	8,
	20
};
static const int ATTACK_COOLDOWNS[3] = {
	160,
	120,
	90
};
static const int SHOOT_COOLDOWNS[3] = {
	35,
	25,
	10
};
static const float SHOT_ANGLES[3] = {
	0.375f,
	0.34375f,
	0.3125f
};

EnemyFirstBossModel * EnemyFirstBossModel_create(DungeonRoomStateModel * stateModel, int rank, Vector2 position, Direction facing, size_t inventoryItemCount, struct InventoryModel_item * inventoryItems, size_t patrolWaypointCount, Vector2 * patrolWaypoints) {
	stemobject_create_implementation(EnemyFirstBossModel, init, stateModel, rank, position, facing, inventoryItemCount, inventoryItems, patrolWaypointCount, patrolWaypoints)
}

static void sharedInit(EnemyFirstBossModel * self) {
	self->dispose = EnemyFirstBossModel_dispose;
	self->copy = EnemyFirstBossModel_copy;
	self->update = EnemyFirstBossModel_update;
	self->protected_ivar(aiUpdate) = EnemyFirstBossModel_aiUpdate;
	self->atRest = EnemyFirstBossModel_atRest;
	self->attack = EnemyFirstBossModel_attack;
	self->protected_ivar(takeDamage) = EnemyFirstBossModel_takeDamage;
	
	self->enemyType = ENEMY_TYPE_FIRST_BOSS;
	self->passiveVisionDistance = 10.0f;
	self->stealthVisionDistance = 10.0f;
	self->knockbackResistance = 0.5f;
	self->radius = 0.5f;
	self->acceleration = 1.0f / 512.0f;
	self->deceleration = 2.0f / 512.0f;
	self->velocityMax = 12.0f / 512.0f;
}

void EnemyFirstBossModel_init(compat_type(EnemyFirstBossModel *) selfPtr, DungeonRoomStateModel * stateModel, int rank, Vector2 position, Direction facing, size_t inventoryItemCount, struct InventoryModel_item * inventoryItems, size_t patrolWaypointCount, Vector2 * patrolWaypoints) {
	EnemyFirstBossModel * self = selfPtr;
	int health;
	
	switch (rank) {
		case 0:
		default:
			health = 3;
			break;
		case 1:
			health = 5;
			break;
		case 2:
			health = 7;
			break;
	}
	EnemyModel_init(self, stateModel, rank, health, position, facing, inventoryItemCount, inventoryItems, patrolWaypointCount, patrolWaypoints);
	sharedInit(self);
	self->shotsLeft = SHOTS_BEFORE_BREAKS[self->rank];
	self->attackCooldown = 0;
	self->shootCooldown = 0;
	self->swordSwingFacingCooldown = 0;
}

void EnemyFirstBossModel_initCopy(compat_type(EnemyFirstBossModel *) selfPtr, compat_type(EnemyFirstBossModel *) originalPtr) {
	EnemyFirstBossModel * self = selfPtr, * original = originalPtr;
	
	EnemyModel_initCopy(self, original);
	sharedInit(self);
	
	self->shotsLeft = original->shotsLeft;
	self->attackCooldown = original->attackCooldown;
	self->shootCooldown = original->shootCooldown;
	self->swordSwingFacing = original->swordSwingFacing;
	self->swordSwingFacingCooldown = original->swordSwingFacingCooldown;
}

void EnemyFirstBossModel_dispose(compat_type(EnemyFirstBossModel *) selfPtr) {
	EnemyFirstBossModel * self = selfPtr;
	
	EnemyModel_dispose(self);
}

compat_type(EnemyFirstBossModel *) EnemyFirstBossModel_copy(compat_type(EnemyFirstBossModel *) selfPtr) {
	stemobject_copy_implementation(EnemyFirstBossModel, initCopy)
}

void EnemyFirstBossModel_update(compat_type(EnemyFirstBossModel *) selfPtr) {
	EnemyFirstBossModel * self = selfPtr;
	
	EnemyModel_update(self);
	
	if (self->aiState == ENEMY_AI_STATE_ATTACK) {
		self->facing = vectorToNearestDirection(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
		self->attack(self);
		self->facing = vectorToNearestDirection(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
	}
	
	if (self->attackCooldown > 0) {
		self->attackCooldown--;
	}
	if (self->shootCooldown > 0) {
		self->shootCooldown--;
	}
	if (self->swordSwingFacingCooldown > 0) {
		self->facing = self->swordSwingFacing;
		self->swordSwingFacingCooldown--;
	}
}

void EnemyFirstBossModel_aiUpdate(compat_type(EnemyFirstBossModel *) selfPtr) {
	EnemyFirstBossModel * self = selfPtr;
	
	if (self->aiState == ENEMY_AI_STATE_ATTACK) {
		self->facing = vectorToNearestDirection(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
	}
	EnemyModel_aiUpdate(self);
	if (self->canSeePlayer) {
		self->setAIState(self, ENEMY_AI_STATE_ATTACK);
	} else {
		self->setAIState(self, ENEMY_AI_STATE_PATROL);
	}
}

bool EnemyFirstBossModel_atRest(compat_type(EnemyFirstBossModel *) selfPtr) {
	EnemyFirstBossModel * self = selfPtr;
	
	return self->attackCooldown <= 0 &&
	       self->shootCooldown <= 0 &&
	       EnemyModel_atRest(self);
}

static Vector2 Vector2_rotate(Vector2 vector, float angle) {
	Vector2 result;
	
	result.x = vector.x * cos(angle) + vector.y * -sin(angle);
	result.y = vector.x * sin(angle) + vector.y * cos(angle);
	return result;
}

void EnemyFirstBossModel_attack(compat_type(EnemyFirstBossModel *) selfPtr) {
	EnemyFirstBossModel * self = selfPtr;
	
	if (self->shotsLeft > 0) {
		if (self->shootCooldown <= 0) {
			ProjectileModel * projectile;
			Vector2 direction, direction2, direction3;
			
			self->shootCooldown = SHOOT_COOLDOWNS[self->rank];
			direction = Vector2_normalized(Vector2_subtract(self->lastKnownPlayerPosition, self->position));
			direction2 = Vector2_rotate(direction, M_PI * SHOT_ANGLES[self->rank]);
			direction3 = Vector2_rotate(direction, M_PI * -SHOT_ANGLES[self->rank]);
			projectile = ProjectileModel_create(self->stateModel, self->position, direction, SHOT_SPEEDS[self->rank], 16.0f, true, DAMAGES[self->rank], KNOCKBACKS[self->rank]);
			projectile->position.x += direction.x * (self->radius - projectile->radius);
			projectile->position.y += direction.y * (self->radius - projectile->radius);
			self->stateModel->addObject(self->stateModel, projectile);
			projectile = ProjectileModel_create(self->stateModel, self->position, direction2, SHOT_SPEEDS[self->rank], 16.0f, true, DAMAGES[self->rank], KNOCKBACKS[self->rank]);
			projectile->position.x += direction.x * (self->radius - projectile->radius);
			projectile->position.y += direction.y * (self->radius - projectile->radius);
			self->stateModel->addObject(self->stateModel, projectile);
			projectile = ProjectileModel_create(self->stateModel, self->position, direction3, SHOT_SPEEDS[self->rank], 16.0f, true, DAMAGES[self->rank], KNOCKBACKS[self->rank]);
			projectile->position.x += direction.x * (self->radius - projectile->radius);
			projectile->position.y += direction.y * (self->radius - projectile->radius);
			self->stateModel->addObject(self->stateModel, projectile);
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(ENEMY_EVENT_SHOOT), NULL);
			self->shotsLeft--;
			if (self->shotsLeft <= 0) {
				self->attackCooldown = ATTACK_COOLDOWNS[self->rank];
			}
		}
		
	} else if (self->attackCooldown <= 0) {
		self->swordSwingFacing = self->facing;
		switch (self->rank) {
			case 0:
			default:
				self->swordSwingFacingCooldown = 40;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, true, 6, 0.1875f, self->radius, 2.75f, 1.0f, 40));
				break;
				
			case 1:
				self->swordSwingFacingCooldown = 30;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, true, 9, 0.25f, self->radius, 3.0f, 1.0f, 30));
				break;
				
			case 2:
				self->swordSwingFacingCooldown = 15;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, true, 15, 0.3125f, self->radius, 3.5f, 1.0f, 15));
				break;
		}
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(ENEMY_EVENT_ATTACK), NULL);
		self->shootCooldown = 60;
		self->shotsLeft = SHOTS_BEFORE_BREAKS[self->rank];
	}
}

void EnemyFirstBossModel_takeDamage(compat_type(EnemyFirstBossModel *) selfPtr, int amount, Direction fromDirection) {
	EnemyFirstBossModel * self = selfPtr;
	
	if (self->shotsLeft == 0) {
		amount = 1;
		self->shootCooldown = 40;
		self->shotsLeft = SHOTS_BEFORE_BREAKS[self->rank];
	} else {
		amount = 0;
	}
	EnemyModel_takeDamage(self, amount, fromDirection);
	self->immunity = 30;
}
