/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/GamepadDetailView.h"
#include "convergence/SharedDefinitions.h"
#include <stddef.h>
#include <stdlib.h>

struct vertex_p2f_c3f {
	GLfloat position[2];
	GLfloat color[3];
};

GamepadDetailView * GamepadDetailView_create(struct Gamepad_device * device) {
	stemobject_create_implementation(GamepadDetailView, init, device)
}

#define BUTTON_VERTICES 12
#define AXIS_VERTICES 16
#define BUTTON_INDEXES 18
#define AXIS_INDEXES 24

#define BUTTON_WIDTH 0.03125f
#define BUTTON_HEIGHT 0.03125f
#define BUTTON_BORDER 0.005859375f
#define BUTTON_PADDING 0.0078125f
#define BUTTON_BORDER_RED 0.0f
#define BUTTON_BORDER_GREEN 0.375f
#define BUTTON_BORDER_BLUE 0.75f
#define BUTTON_FILL_UP_RED 0.0f
#define BUTTON_FILL_UP_GREEN 0.0f
#define BUTTON_FILL_UP_BLUE 0.0f
#define BUTTON_FILL_DOWN_RED 1.0f
#define BUTTON_FILL_DOWN_GREEN 0.5f
#define BUTTON_FILL_DOWN_BLUE 0.0f
#define AXIS_WIDTH 0.09375f
#define AXIS_HEIGHT 0.03125f
#define AXIS_BORDER 0.005859375f
#define AXIS_PADDING 0.0078125f
#define AXIS_BORDER_RED 0.0f
#define AXIS_BORDER_GREEN 0.375f
#define AXIS_BORDER_BLUE 0.75f
#define AXIS_FILL_RED 0.0f
#define AXIS_FILL_GREEN 0.0f
#define AXIS_FILL_BLUE 0.0f
#define AXIS_POSITION_RED 1.0f
#define AXIS_POSITION_GREEN 0.5f
#define AXIS_POSITION_BLUE 0.0f
#define AXIS_POSITION_WIDTH 0.005859375f

void GamepadDetailView_init(compat_type(GamepadDetailView *) selfPtr, struct Gamepad_device * device) {
	GamepadDetailView * self = selfPtr;
	GLushort * indexes;
	size_t buttonIndex, axisIndex, quadIndex;
	
	StemObject_init(self);
	self->dispose = GamepadDetailView_dispose;
	self->mappableElementAtPosition = GamepadDetailView_mappableElementAtPosition;
	self->draw = GamepadDetailView_draw;
	
	self->device = device;
	glGenBuffers(1, &self->vboID);
	glGenBuffers(1, &self->eboID);
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_c3f) * device->numButtons * BUTTON_VERTICES * device->numAxes * AXIS_VERTICES, NULL, GL_STREAM_DRAW);
	indexes = malloc(sizeof(GLushort) * (self->device->numButtons * BUTTON_INDEXES + self->device->numAxes * AXIS_INDEXES));
	for (buttonIndex = 0; buttonIndex < self->device->numButtons; buttonIndex++) {
		for (quadIndex = 0; quadIndex < 3; quadIndex++) {
			indexes[buttonIndex * BUTTON_INDEXES + quadIndex * 6 + 0]  = buttonIndex * BUTTON_VERTICES + quadIndex * 4 + 0;
			indexes[buttonIndex * BUTTON_INDEXES + quadIndex * 6 + 1]  = buttonIndex * BUTTON_VERTICES + quadIndex * 4 + 1;
			indexes[buttonIndex * BUTTON_INDEXES + quadIndex * 6 + 2]  = buttonIndex * BUTTON_VERTICES + quadIndex * 4 + 2;
			indexes[buttonIndex * BUTTON_INDEXES + quadIndex * 6 + 3]  = buttonIndex * BUTTON_VERTICES + quadIndex * 4 + 2;
			indexes[buttonIndex * BUTTON_INDEXES + quadIndex * 6 + 4]  = buttonIndex * BUTTON_VERTICES + quadIndex * 4 + 3;
			indexes[buttonIndex * BUTTON_INDEXES + quadIndex * 6 + 5]  = buttonIndex * BUTTON_VERTICES + quadIndex * 4 + 0;
		}
	}
	for (axisIndex = 0; axisIndex < self->device->numAxes; axisIndex++) {
		for (quadIndex = 0; quadIndex < 4; quadIndex++) {
			indexes[buttonIndex * BUTTON_INDEXES + axisIndex * AXIS_INDEXES + quadIndex * 6 + 0]  = buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + quadIndex * 4 + 0;
			indexes[buttonIndex * BUTTON_INDEXES + axisIndex * AXIS_INDEXES + quadIndex * 6 + 1]  = buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + quadIndex * 4 + 1;
			indexes[buttonIndex * BUTTON_INDEXES + axisIndex * AXIS_INDEXES + quadIndex * 6 + 2]  = buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + quadIndex * 4 + 2;
			indexes[buttonIndex * BUTTON_INDEXES + axisIndex * AXIS_INDEXES + quadIndex * 6 + 3]  = buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + quadIndex * 4 + 2;
			indexes[buttonIndex * BUTTON_INDEXES + axisIndex * AXIS_INDEXES + quadIndex * 6 + 4]  = buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + quadIndex * 4 + 3;
			indexes[buttonIndex * BUTTON_INDEXES + axisIndex * AXIS_INDEXES + quadIndex * 6 + 5]  = buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + quadIndex * 4 + 0;
		}
	}
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * (self->device->numButtons * BUTTON_INDEXES + self->device->numAxes * AXIS_INDEXES), indexes, GL_STATIC_DRAW);
	free(indexes);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	self->width = BUTTON_PADDING + (BUTTON_WIDTH + BUTTON_PADDING) * self->device->numButtons + (AXIS_WIDTH + AXIS_PADDING) * self->device->numAxes;
}

void GamepadDetailView_dispose(compat_type(GamepadDetailView *) selfPtr) {
	GamepadDetailView * self = selfPtr;
	
	glDeleteBuffers(1, &self->vboID);
	glDeleteBuffers(1, &self->eboID);
	StemObject_dispose(self);
}

struct GamepadDetailView_mappableElement GamepadDetailView_mappableElementAtPosition(compat_type(GamepadDetailView *) selfPtr, float position) {
	GamepadDetailView * self = selfPtr;
	struct GamepadDetailView_mappableElement result;
	
	if (position < 0.0f) {
		if (self->device->numButtons > 0) {
			result.axis = false;
			result.u.button.buttonID = 0;
		} else {
			result.axis = true;
			result.u.axis.axisID = 0;
			result.u.axis.direction = -1;
		}
		
	} else if (position < (BUTTON_WIDTH + BUTTON_PADDING) * self->device->numButtons) {
		result.axis = false;
		result.u.button.buttonID = max(0, min(self->device->numButtons - 1, position / (BUTTON_WIDTH + BUTTON_PADDING)));
		
	} else if (position < (BUTTON_WIDTH + BUTTON_PADDING) * self->device->numButtons + (AXIS_WIDTH + AXIS_PADDING) * self->device->numAxes) {
		result.axis = true;
		position -= (BUTTON_WIDTH + BUTTON_PADDING) * self->device->numButtons;
		position /= AXIS_WIDTH + AXIS_PADDING;
		result.u.axis.axisID = max(0, min(self->device->numAxes - 1, position));
		result.u.axis.direction = fmod(position, 1.0f) < 0.5f ? -1 : 1;
		
	} else {
		if (self->device->numButtons > 0) {
			result.axis = false;
			result.u.button.buttonID = self->device->numButtons - 1;
		} else {
			result.axis = true;
			result.u.axis.axisID = self->device->numAxes - 1;
			result.u.axis.direction = 1;
		}
	}
	return result;
}

void GamepadDetailView_draw(compat_type(GamepadDetailView *) selfPtr, Matrix modelviewMatrix) {
	GamepadDetailView * self = selfPtr;
	struct vertex_p2f_c3f * vertices;
	size_t buttonIndex, axisIndex;
	struct vertex_p2f_c3f vertex;
	float offsetX, offsetY;
	
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	vertices = glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
	
	offsetX = BUTTON_PADDING;
	offsetY = 0.0f;
	for (buttonIndex = 0; buttonIndex < self->device->numButtons; buttonIndex++) {
		vertex.color[0] = BUTTON_BORDER_RED;
		vertex.color[1] = BUTTON_BORDER_GREEN;
		vertex.color[2] = BUTTON_BORDER_BLUE;
		vertex.position[0] = offsetX;
		vertex.position[1] = offsetY;
		vertices[buttonIndex * BUTTON_VERTICES + 0] = vertex;
		vertex.position[0] = offsetX + BUTTON_WIDTH;
		vertices[buttonIndex * BUTTON_VERTICES + 1] = vertex;
		vertex.position[1] = offsetY + BUTTON_HEIGHT;
		vertices[buttonIndex * BUTTON_VERTICES + 2] = vertex;
		vertex.position[0] = offsetX;
		vertices[buttonIndex * BUTTON_VERTICES + 3] = vertex;
		vertex.position[0] = offsetX + BUTTON_BORDER;
		vertex.position[1] = offsetY + BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 4] = vertex;
		vertex.position[0] = offsetX + BUTTON_WIDTH - BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 5] = vertex;
		vertex.position[1] = offsetY + BUTTON_HEIGHT - BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 6] = vertex;
		vertex.position[0] = offsetX + BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 7] = vertex;
		
		if (self->device->buttonStates[buttonIndex]) {
			vertex.color[0] = BUTTON_FILL_DOWN_RED;
			vertex.color[1] = BUTTON_FILL_DOWN_GREEN;
			vertex.color[2] = BUTTON_FILL_DOWN_BLUE;
		} else {
			vertex.color[0] = BUTTON_FILL_UP_RED;
			vertex.color[1] = BUTTON_FILL_UP_GREEN;
			vertex.color[2] = BUTTON_FILL_UP_BLUE;
		}
		vertex.position[0] = offsetX + BUTTON_BORDER;
		vertex.position[1] = offsetY + BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 8] = vertex;
		vertex.position[0] = offsetX + BUTTON_WIDTH - BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 9] = vertex;
		vertex.position[1] = offsetY + BUTTON_HEIGHT - BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 10] = vertex;
		vertex.position[0] = offsetX + BUTTON_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + 11] = vertex;
		
		offsetX += BUTTON_WIDTH + BUTTON_PADDING;
	}
	for (axisIndex = 0; axisIndex < self->device->numAxes; axisIndex++) {
		vertex.color[0] = AXIS_BORDER_RED;
		vertex.color[1] = AXIS_BORDER_GREEN;
		vertex.color[2] = AXIS_BORDER_BLUE;
		vertex.position[0] = offsetX;
		vertex.position[1] = offsetY;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 0] = vertex;
		vertex.position[0] = offsetX + AXIS_WIDTH;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 1] = vertex;
		vertex.position[1] = offsetY + AXIS_HEIGHT;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 2] = vertex;
		vertex.position[0] = offsetX;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 3] = vertex;
		vertex.position[0] = offsetX + AXIS_BORDER;
		vertex.position[1] = offsetY + AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 4] = vertex;
		vertex.position[0] = offsetX + AXIS_WIDTH - AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 5] = vertex;
		vertex.position[1] = offsetY + AXIS_HEIGHT - AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 6] = vertex;
		vertex.position[0] = offsetX + AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 7] = vertex;
		
		vertex.color[0] = AXIS_FILL_RED;
		vertex.color[1] = AXIS_FILL_GREEN;
		vertex.color[2] = AXIS_FILL_BLUE;
		vertex.position[0] = offsetX + AXIS_BORDER;
		vertex.position[1] = offsetY + AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 8] = vertex;
		vertex.position[0] = offsetX + AXIS_WIDTH - AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 9] = vertex;
		vertex.position[1] = offsetY + AXIS_HEIGHT - AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 10] = vertex;
		vertex.position[0] = offsetX + AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 11] = vertex;
		
		vertex.color[0] = AXIS_POSITION_RED;
		vertex.color[1] = AXIS_POSITION_GREEN;
		vertex.color[2] = AXIS_POSITION_BLUE;
		vertex.position[0] = offsetX + (AXIS_WIDTH - AXIS_POSITION_WIDTH) * 0.5f + (AXIS_WIDTH - AXIS_BORDER * 2 - AXIS_POSITION_WIDTH) * self->device->axisStates[axisIndex] * 0.5f;
		vertex.position[1] = offsetY + AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 12] = vertex;
		vertex.position[0] = offsetX + (AXIS_WIDTH + AXIS_POSITION_WIDTH) * 0.5f + (AXIS_WIDTH - AXIS_BORDER * 2 - AXIS_POSITION_WIDTH) * self->device->axisStates[axisIndex] * 0.5f;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 13] = vertex;
		vertex.position[1] = offsetY + AXIS_HEIGHT - AXIS_BORDER;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 14] = vertex;
		vertex.position[0] = offsetX + (AXIS_WIDTH - AXIS_POSITION_WIDTH) * 0.5f + (AXIS_WIDTH - AXIS_BORDER * 2 - AXIS_POSITION_WIDTH) * self->device->axisStates[axisIndex] * 0.5f;
		vertices[buttonIndex * BUTTON_VERTICES + axisIndex * AXIS_VERTICES + 15] = vertex;
		
		offsetX += AXIS_WIDTH + AXIS_PADDING;
	}
	
	glUnmapBuffer(GL_ARRAY_BUFFER);
	
	glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_c3f), (void *) offsetof(struct vertex_p2f_c3f, position));
	glColorPointer(3, GL_FLOAT, sizeof(struct vertex_p2f_c3f), (void *) offsetof(struct vertex_p2f_c3f, color));
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	glDrawElements(GL_TRIANGLES, buttonIndex * BUTTON_INDEXES + axisIndex * AXIS_INDEXES, GL_UNSIGNED_SHORT, 0);
	
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
