/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "3dmath/Matrix.h"
#include "convergence/DeathEffect.h"
#include "convergence/DungeonDoorView.h"
#include "convergence/DungeonKeyView.h"
#include "convergence/EnemyModel.h"
#include "convergence/EnemyView.h"
#include "convergence/GameplayStateView.h"
#include "convergence/GraphicsConstants.h"
#include "convergence/HealPowerupView.h"
#include "convergence/MovingEntityView.h"
#include "convergence/PlayerModel.h"
#include "convergence/PlayerView.h"
#include "convergence/ProjectileView.h"
#include "convergence/RainbowTrailEffect.h"
#include "convergence/RankUpPowerupView.h"
#include "convergence/ShellStateGlobals.h"
#include "convergence/SparkleEffect.h"
#include "convergence/StaircaseView.h"
#include "convergence/SwordSwingView.h"
#include "shell/Shell.h"
#include "utilities/Random.h"
#include <stddef.h>

struct vertex_p2f {
	GLfloat position[2];
};

GameplayStateView * GameplayStateView_create(GameStateModel * gameStateModel, GameplayCamera * camera, ResourceManager * resourceManager) {
	stemobject_create_implementation(GameplayStateView, init, gameStateModel, camera, resourceManager)
}

static void addView(GameplayStateView * self, compat_type(GameObjectView *) view) {
	size_t viewIndex, viewIndex2;
	
	if (self->objectViewListSize <= self->objectViewCount) {
		self->objectViewListSize = self->objectViewListSize == 0 ? 1 : self->objectViewListSize * 2;
		self->objectViews = realloc(self->objectViews, sizeof(GameObjectView *) * self->objectViewListSize);
	}
	for (viewIndex = 0; viewIndex < self->objectViewCount; viewIndex++) {
		if (self->objectViews[viewIndex]->sortOrder > ((GameObjectView *) view)->sortOrder) {
			for (viewIndex2 = self->objectViewCount; viewIndex2 > viewIndex; viewIndex2--) {
				self->objectViews[viewIndex2] = self->objectViews[viewIndex2 - 1];
			}
			break;
		}
	}
	self->objectViews[viewIndex] = view;
	self->objectViewCount++;
}

static bool dungeonComplete(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayStateView * self = context;
	SparkleEffect * sparkle;
	
	self->dungeonComplete = true;
	self->dungeonCompletionTime = self->currentTime;
	sparkle = SparkleEffect_create(self->camera->position, 6.0f, NULL);
	sparkle->sortOrder++;
	addView(self, sparkle);
	return true;
}

static bool playerDash(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayStateView * self = context;
	
	addView(self, RainbowTrailEffect_create(sender, self->resourceManager));
	return true;
}

static bool playerRankUp(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayStateView * self = context;
	PlayerModel * playerModel = sender;
	
	addView(self, SparkleEffect_create(playerModel->position, 2.0f, sender));
	return true;
}

static bool playerDeath(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayStateView * self = context;
	PlayerModel * playerModel = sender;
	
	addView(self, DeathEffect_create(playerModel->position, self->resourceManager));
	playerModel->eventDispatcher->unregisterForEvent(playerModel->eventDispatcher, Atom_fromString(PLAYER_EVENT_DASH), playerDash, self);
	playerModel->eventDispatcher->unregisterForEvent(playerModel->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), playerRankUp, self);
	playerModel->eventDispatcher->unregisterForEvent(playerModel->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
	return true;
}

static bool enemyDeath(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayStateView * self = context;
	EnemyModel * enemyModel = sender;
	
	addView(self, DeathEffect_create(enemyModel->position, self->resourceManager));
	enemyModel->eventDispatcher->unregisterForEvent(enemyModel->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), enemyDeath, self);
	return true;
}

static bool objectAdded(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayStateView * self = context;
	GameObjectModel * object = eventData;
	compat_type(GameObjectView *) view = NULL;
	
	switch (object->type) {
		case GAME_OBJECT_TYPE_PLAYER:
			object->eventDispatcher->registerForEvent(object->eventDispatcher, Atom_fromString(PLAYER_EVENT_DASH), playerDash, self);
			object->eventDispatcher->registerForEvent(object->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), playerRankUp, self);
			object->eventDispatcher->registerForEvent(object->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
			view = PlayerView_create(object);
			break;
		case GAME_OBJECT_TYPE_ENEMY:
			object->eventDispatcher->registerForEvent(object->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), enemyDeath, self);
			view = EnemyView_create(object);
			break;
		case GAME_OBJECT_TYPE_SWORD_SWING:
			view = SwordSwingView_create(object);
			break;
		case GAME_OBJECT_TYPE_PROJECTILE:
			view = ProjectileView_create(object);
			break;
		case GAME_OBJECT_TYPE_HEAL_POWERUP:
			view = HealPowerupView_create(object, self->resourceManager);
			break;
		case GAME_OBJECT_TYPE_DOOR:
			view = DungeonDoorView_create(object);
			break;
		case GAME_OBJECT_TYPE_KEY:
			view = DungeonKeyView_create(object, self->resourceManager);
			break;
		case GAME_OBJECT_TYPE_RANK_UP_POWERUP:
			view = RankUpPowerupView_create(object, self->resourceManager);
			break;
		case GAME_OBJECT_TYPE_STAIRCASE:
			view = StaircaseView_create(object, self->resourceManager);
			break;
		default:
			view = NULL;
			break;
	}
	if (view != NULL) {
		addView(self, view);
		return true;
	}
	return false;
}

static bool objectRemoved(void * sender, Atom eventID, void * eventData, void * context) {
	GameplayStateView * self = context;
	size_t objectIndex;
	
	for (objectIndex = 0; objectIndex < self->objectViewCount; objectIndex++) {
		if (self->objectViews[objectIndex]->model == eventData) {
			if (self->objectViews[objectIndex]->model->type == GAME_OBJECT_TYPE_ENEMY) {
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), enemyDeath, self);
			} else if (self->objectViews[objectIndex]->model->type == GAME_OBJECT_TYPE_PLAYER) {
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DASH), playerDash, self);
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), playerRankUp, self);
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
			}
			self->objectViews[objectIndex]->dispose(self->objectViews[objectIndex]);
			self->objectViewCount--;
			for (; objectIndex < self->objectViewCount; objectIndex++) {
				self->objectViews[objectIndex] = self->objectViews[objectIndex + 1];
			}
			return true;
		}
	}
	return false;
}

void GameplayStateView_init(compat_type(GameplayStateView *) selfPtr, GameStateModel * gameStateModel, GameplayCamera * camera, ResourceManager * resourceManager) {
	GameplayStateView * self = selfPtr;
	size_t objectIndex;
	struct vertex_p2f vertices[4] = {{{-1.0f, -1.0f}}, {{1.0f, -1.0f}}, {{1.0f, 1.0f}}, {{-1.0f, 1.0f}}};
	GLubyte indexes[4] = {0, 1, 2, 3};
	
	StemObject_init(self);
	self->dispose = GameplayStateView_dispose;
	self->playFadeEffect = GameplayStateView_playFadeEffect;
	self->draw = GameplayStateView_draw;
	
	self->gameStateModel = gameStateModel;
	self->gameStateModel->eventDispatcher->registerForEvent(self->gameStateModel->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_DUNGEON_COMPLETE), dungeonComplete, self);
	self->camera = camera;
	self->roomState = self->gameStateModel->activeRoom;
	self->roomState->eventDispatcher->registerForEvent(self->roomState->eventDispatcher, Atom_fromString(DUNGEON_ROOM_STATE_EVENT_OBJECT_ADDED), objectAdded, self);
	self->roomState->eventDispatcher->registerForEvent(self->roomState->eventDispatcher, Atom_fromString(DUNGEON_ROOM_STATE_EVENT_OBJECT_REMOVED), objectRemoved, self);
	self->resourceManager = resourceManager;
	
	self->roomView = DungeonRoomView_create(self->roomState, resourceManager);
	self->objectViewCount = 0;
	self->objectViewListSize = self->roomState->objectCount;
	self->objectViews = malloc(sizeof(GameObjectView *) * self->objectViewListSize);
	for (objectIndex = 0; objectIndex < self->roomState->objectCount; objectIndex++) {
		objectAdded(NULL, DUNGEON_ROOM_STATE_EVENT_OBJECT_ADDED, self->roomState->objects[objectIndex], self);
	}
	
	self->dungeonComplete = false;
	self->currentTime = 0.0;
	
	self->playingFadeEffect = false;
	glGenBuffers(1, &self->vboID);
	glGenBuffers(1, &self->eboID);
	glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f) * 4, vertices, GL_STATIC_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLubyte) * 4, indexes, GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}

void GameplayStateView_dispose(compat_type(GameplayStateView *) selfPtr) {
	GameplayStateView * self = selfPtr;
	size_t objectIndex;
	
	self->gameStateModel->eventDispatcher->unregisterForEvent(self->gameStateModel->eventDispatcher, Atom_fromString(GAME_STATE_EVENT_DUNGEON_COMPLETE), dungeonComplete, self);
	self->camera->dispose(self->camera);
	self->roomState->eventDispatcher->unregisterForEvent(self->roomState->eventDispatcher, Atom_fromString(DUNGEON_ROOM_STATE_EVENT_OBJECT_ADDED), objectAdded, self);
	self->roomState->eventDispatcher->unregisterForEvent(self->roomState->eventDispatcher, Atom_fromString(DUNGEON_ROOM_STATE_EVENT_OBJECT_REMOVED), objectRemoved, self);
	self->roomView->dispose(self->roomView);
	for (objectIndex = 0; objectIndex < self->objectViewCount; objectIndex++) {
		if (self->objectViews[objectIndex]->model != NULL) {
			if (self->objectViews[objectIndex]->model->type == GAME_OBJECT_TYPE_ENEMY) {
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(ENEMY_EVENT_DEATH), enemyDeath, self);
			} else if (self->objectViews[objectIndex]->model->type == GAME_OBJECT_TYPE_PLAYER) {
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DASH), playerDash, self);
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), playerRankUp, self);
				self->objectViews[objectIndex]->model->eventDispatcher->unregisterForEvent(self->objectViews[objectIndex]->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), playerDeath, self);
			}
		}
		self->objectViews[objectIndex]->dispose(self->objectViews[objectIndex]);
	}
	free(self->objectViews);
	glDeleteBuffers(1, &self->vboID);
	glDeleteBuffers(1, &self->eboID);
	StemObject_dispose(self);
}

void GameplayStateView_playFadeEffect(compat_type(GameplayStateView *) selfPtr, uint32_t startColor, uint32_t endColor, double duration) {
	GameplayStateView * self = selfPtr;
	
	self->playingFadeEffect = true;
	self->fadeEffectStartColor = startColor;
	self->fadeEffectEndColor = endColor;
	self->fadeEffectDuration = duration;
	self->fadeEffectStartTime = self->currentTime;
}

void GameplayStateView_draw(compat_type(GameplayStateView *) selfPtr, double deltaTime) {
	GameplayStateView * self = selfPtr;
	Matrix modelviewMatrix, projectionMatrix;
	float ratio;
	size_t objectIndex, objectIndex2;
	
	self->currentTime += deltaTime;
	
	ratio = (float) g_windowWidth / (float) g_windowHeight;
	if (self->objectViewCount > 0) {
		objectIndex = self->objectViewCount;
		do {
			objectIndex--;
			if (self->objectViews[objectIndex]->markedForRemoval) {
				self->objectViews[objectIndex]->dispose(self->objectViews[objectIndex]);
				self->objectViewCount--;
				for (objectIndex2 = objectIndex; objectIndex2 < self->objectViewCount; objectIndex2++) {
					self->objectViews[objectIndex2] = self->objectViews[objectIndex2 + 1];
				}
			}
		} while (objectIndex > 0);
	}
	
	if (self->dungeonComplete) {
		addView(self, SparkleEffect_create(Vector2_init(self->camera->position.x + frand(ratio * 8.0f), self->camera->position.y + frand(8.0f)), ufrand(2.5f) + 0.5f, NULL));
		addView(self, SparkleEffect_create(Vector2_init(self->camera->position.x + frand(ratio * 8.0f), self->camera->position.y + frand(8.0f)), ufrand(2.5f) + 0.5f, NULL));
		addView(self, SparkleEffect_create(Vector2_init(self->camera->position.x + frand(ratio * 8.0f), self->camera->position.y + frand(8.0f)), ufrand(2.5f) + 0.5f, NULL));
	}
	
	glClear(GL_COLOR_BUFFER_BIT);
	projectionMatrix = Matrix_ortho(Matrix_identity(), -ratio, ratio, -1.0f, 1.0f, -1.0f, 1.0f);
	modelviewMatrix = Matrix_identity();
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projectionMatrix.m);
	
	self->camera->update(self->camera);
	modelviewMatrix = self->camera->getMatrix(self->camera);
	
	self->roomView->draw(self->roomView, modelviewMatrix);
	for (objectIndex = 0; objectIndex < self->objectViewCount; objectIndex++) {
		self->objectViews[objectIndex]->draw(self->objectViews[objectIndex], modelviewMatrix, deltaTime);
	}
	
	if (self->playingFadeEffect) {
		float step;
		unsigned char red, green, blue, alpha;
		
		step = (self->currentTime - self->fadeEffectStartTime) / self->fadeEffectDuration;
		if (step > 1.0f) {
			step = 1.0f;
		}
		red = (self->fadeEffectStartColor >> 24) * (1.0f - step) + (self->fadeEffectEndColor >> 24) * step;
		green = (self->fadeEffectStartColor >> 16 & 0xFF) * (1.0f - step) + (self->fadeEffectEndColor >> 16 & 0xFF) * step;
		blue = (self->fadeEffectStartColor >> 8 & 0xFF) * (1.0f - step) + (self->fadeEffectEndColor >> 8 & 0xFF) * step;
		alpha = (self->fadeEffectStartColor & 0xFF) * (1.0f - step) + (self->fadeEffectEndColor & 0xFF) * step;
		glBindBuffer(GL_ARRAY_BUFFER, self->vboID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->eboID);
		glEnable(GL_BLEND);
		glEnableClientState(GL_VERTEX_ARRAY);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glScalef(ratio, 1.0f, 1.0f);
		glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f), (void *) offsetof(struct vertex_p2f, position));
		glColor4ub(red, green, blue, alpha);
		glDrawElements(GL_TRIANGLE_FAN, 4, GL_UNSIGNED_BYTE, 0);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisable(GL_BLEND);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	}
}
