/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/DungeonDoorModel.h"
#include "convergence/DungeonKeyModel.h"
#include "convergence/HealPowerupModel.h"
#include "convergence/PlayerModel.h"
#include "convergence/ProjectileModel.h"
#include "convergence/RoomExitModel.h"
#include "convergence/SwordSwingModel.h"
#include <stdlib.h>

PlayerModel * PlayerModel_create(DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	stemobject_create_implementation(PlayerModel, init, stateModel, position, facing)
}

static void sharedInit(PlayerModel * self) {
	self->dispose = PlayerModel_dispose;
	self->copy = PlayerModel_copy;
	self->update = PlayerModel_update;
	self->atRest = PlayerModel_atRest;
	self->wouldCollide = PlayerModel_wouldCollide;
	self->protected_ivar(collisionInternal) = PlayerModel_collisionInternal;
	self->startAttack = PlayerModel_startAttack;
	self->stopAttack = PlayerModel_stopAttack;
	self->startSpecialAction = PlayerModel_startSpecialAction;
	self->stopSpecialAction = PlayerModel_stopSpecialAction;
	self->spotted = PlayerModel_spotted;
	self->protected_ivar(knockedBack) = PlayerModel_knockedBack;
	self->protected_ivar(takeDamage) = PlayerModel_takeDamage;
	
	self->type = GAME_OBJECT_TYPE_PLAYER;
	self->acceleration = 0.00390625f;
	self->deceleration = 0.0078125f;
	self->velocityMax = 0.046875f;
}

void PlayerModel_init(compat_type(PlayerModel *) selfPtr, DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	PlayerModel * self = selfPtr;
	
	MovingEntityModel_init(self, stateModel);
	sharedInit(self);
	
	self->inventory = InventoryModel_create();
	self->rank = 0;
	self->maxHealth = 20;
	self->health = 20;
	self->hitCooldown = 0;
	self->position = position;
	self->facing = facing;
	self->exited = false;
}

void PlayerModel_initCopy(compat_type(PlayerModel *) selfPtr, compat_type(PlayerModel *) originalPtr) {
	PlayerModel * self = selfPtr, * original = originalPtr;
	
	MovingEntityModel_initCopy(self, original);
	sharedInit(self);
	
	self->rank = original->rank;
	self->inventory = original->inventory->copy(original->inventory);
	self->maxHealth = original->maxHealth;
	self->health = original->health;
	self->hitCooldown = original->hitCooldown;
	self->exited = original->exited;
}

void PlayerModel_dispose(compat_type(PlayerModel *) selfPtr) {
	PlayerModel * self = selfPtr;
	
	self->inventory->dispose(self->inventory);
	StemObject_dispose(self);
}

compat_type(PlayerModel *) PlayerModel_copy(compat_type(PlayerModel *) selfPtr) {
	stemobject_copy_implementation(PlayerModel, initCopy)
}

void PlayerModel_update(compat_type(PlayerModel *) selfPtr) {
	PlayerModel * self = selfPtr;
	
	MovingEntityModel_update(self);
	
	self->exited = false;
	if (self->hitCooldown > 0) {
		self->hitCooldown--;
	}
}

bool PlayerModel_atRest(compat_type(PlayerModel *) selfPtr) {
	PlayerModel * self = selfPtr;
	return self->hitCooldown == 0 &&
	       MovingEntityModel_atRest(self);
}

bool PlayerModel_wouldCollide(compat_type(PlayerModel *) selfPtr, compat_type(GameObjectModel *) objectPtr) {
	PlayerModel * self = selfPtr;
	
	if (self->markedForRemoval) {
		return false;
	}
	switch (((GameObjectModel *) objectPtr)->type) {
		case GAME_OBJECT_TYPE_ENEMY:
		case GAME_OBJECT_TYPE_PLAYER:
		case GAME_OBJECT_TYPE_DOOR:
		case GAME_OBJECT_TYPE_HEAL_POWERUP:
		case GAME_OBJECT_TYPE_KEY:
			return true;
			
		case GAME_OBJECT_TYPE_RANK_UP_POWERUP:
			return self->rank < 2;
			
		case GAME_OBJECT_TYPE_SWORD_SWING:
			return self->hitCooldown == 0 && ((SwordSwingModel *) objectPtr)->hostile;
			
		case GAME_OBJECT_TYPE_PROJECTILE:
			return self->hitCooldown == 0 && ((ProjectileModel *) objectPtr)->hostile;
			
		case GAME_OBJECT_TYPE_ROOM_EXIT:
			return !self->exited;
			
		default:
			return false;
	}
}

void PlayerModel_collisionInternal(compat_type(PlayerModel *) selfPtr, compat_type(GameObjectModel *) objectPtr, const struct collision collision, float time) {
	PlayerModel * self = selfPtr;
	GameObjectModel * object = objectPtr;
	
	MovingEntityModel_collisionInternal(self, objectPtr, collision, time);
	if (object != NULL) {
		if (object->type == GAME_OBJECT_TYPE_SWORD_SWING) {
			SwordSwingModel * swing = objectPtr;
			
			self->protected_ivar(knockedBack)(self, Vector2_init(collision.normal.x * swing->knockback - self->velocity.x, collision.normal.y * swing->knockback - self->velocity.y));
			self->protected_ivar(takeDamage)(self, swing->damage, vectorToNearestDirection(Vector2_init(-collision.normal.x, -collision.normal.y)));
			
		} else if (object->type == GAME_OBJECT_TYPE_PROJECTILE) {
			ProjectileModel * projectile = objectPtr;
			
			self->protected_ivar(knockedBack)(self, Vector2_init(projectile->direction.x * projectile->knockback - self->velocity.x, projectile->direction.y * projectile->knockback - self->velocity.y));
			self->protected_ivar(takeDamage)(self, projectile->damage, (vectorToNearestDirection(projectile->direction) + 4) % 8);
			
		} else if (object->type == GAME_OBJECT_TYPE_HEAL_POWERUP) {
			HealPowerupModel * powerup = objectPtr;
			
			self->health = min(self->health + powerup->amount, self->maxHealth);
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_HEALED), NULL);
			
		} else if (object->type == GAME_OBJECT_TYPE_KEY) {
			DungeonKeyModel * key = objectPtr;
			
			self->inventory->addItem(self->inventory, INVENTORY_ITEM_TYPE_KEY, key->keyID);
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_GOT_KEY), NULL);
			
		} else if (object->type == GAME_OBJECT_TYPE_RANK_UP_POWERUP) {
			self->rank++;
			self->maxHealth += 5 + 5 * self->rank;
			self->health += 5 + 5 * self->rank;
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), NULL);
			
		} else if (object->type == GAME_OBJECT_TYPE_DOOR) {
			DungeonDoorModel * door = objectPtr;
			
			self->inventory->useKey(self->inventory, door->keyID);
			
		} else if (object->type == GAME_OBJECT_TYPE_ROOM_EXIT) {
			RoomExitModel * exit = objectPtr;
			
			self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_EXITED), (void *) exit->name);
			self->exited = true;
		}
	}
}

void PlayerModel_startAttack(compat_type(PlayerModel *) selfPtr) {
}

void PlayerModel_stopAttack(compat_type(PlayerModel *) selfPtr) {
}

void PlayerModel_startSpecialAction(compat_type(PlayerModel *) selfPtr) {
}

void PlayerModel_stopSpecialAction(compat_type(PlayerModel *) selfPtr) {
}

void PlayerModel_spotted(compat_type(PlayerModel *) selfPtr, compat_type(EnemyModel *) enemy) {
}

void PlayerModel_knockedBack(compat_type(PlayerModel *) selfPtr, Vector2 knockback) {
	PlayerModel * self = selfPtr;
	
	self->velocity.x += knockback.x;
	self->velocity.y += knockback.y;
}

void PlayerModel_takeDamage(compat_type(PlayerModel *) selfPtr, int amount, Direction fromDirection) {
	PlayerModel * self = selfPtr;
	
	self->hitCooldown = 20;
	self->health -= amount;
	if (self->health <= 0) {
		self->markedForRemoval = true;
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), NULL);
	} else {
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_HIT), &amount);
	}
}
