/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/PlayerSpyModel.h"
#include "convergence/ProjectileModel.h"
#include "convergence/SwordSwingModel.h"

#define COMBAT_STEALTH_COOLDOWN 40

static const float STEALTH_MAX_VELOCITIES[3] = {
	10.0f / 256.0f,
	11.5f / 256.0f,
	13.0f / 256.0f,
};
static const float MAX_VELOCITIES[3] = {
	15.0f / 256.0f,
	16.0f / 256.0f,
	17.0f / 256.0f,
};

PlayerSpyModel * PlayerSpyModel_create(DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	stemobject_create_implementation(PlayerSpyModel, init, stateModel, position, facing)
}

static bool rankUp(void * sender, Atom eventID, void * eventData, void * context) {
	PlayerSpyModel * self = context;
	
	self->velocityMax = (self->stealth ? STEALTH_MAX_VELOCITIES : MAX_VELOCITIES)[self->rank];
	return true;
}

static void sharedInit(PlayerSpyModel * self) {
	self->dispose = PlayerSpyModel_dispose;
	self->copy = PlayerSpyModel_copy;
	self->update = PlayerSpyModel_update;
	self->atRest = PlayerSpyModel_atRest;
	self->startAttack = PlayerSpyModel_startAttack;
	self->startSpecialAction = PlayerSpyModel_startSpecialAction;
	self->spotted = PlayerSpyModel_spotted;
	self->protected_ivar(takeDamage) = PlayerSpyModel_takeDamage;
	self->enterStealth = PlayerSpyModel_enterStealth;
	self->exitStealth = PlayerSpyModel_exitStealth;
	
	self->playerClass = PLAYER_CLASS_SPY;
	self->acceleration = 1.0f / 256.0f;
	self->deceleration = 2.0f / 256.0f;
	self->velocityMax = MAX_VELOCITIES[self->rank];
	
	self->eventDispatcher->registerForEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), rankUp, self);
}

void PlayerSpyModel_init(compat_type(PlayerSpyModel *) selfPtr, DungeonRoomStateModel * stateModel, Vector2 position, Direction facing) {
	PlayerSpyModel * self = selfPtr;
	
	PlayerModel_init(self, stateModel, position, facing);
	sharedInit(self);
	self->attackCooldown = 0;
	self->stealth = false;
	self->stealthCooldown = 0;
}

void PlayerSpyModel_initCopy(compat_type(PlayerSpyModel *) selfPtr, compat_type(PlayerSpyModel *) originalPtr) {
	PlayerSpyModel * self = selfPtr, * original = originalPtr;
	
	PlayerModel_initCopy(self, original);
	sharedInit(self);
	self->attackCooldown = original->attackCooldown;
	self->stealth = original->stealth;
	self->stealthCooldown = original->stealthCooldown;
	self->swordSwingFacing = original->swordSwingFacing;
	self->velocityMax = original->velocityMax;
}

void PlayerSpyModel_dispose(compat_type(PlayerSpyModel *) selfPtr) {
	PlayerSpyModel * self = selfPtr;
	
	PlayerModel_dispose(self);
}

compat_type(PlayerSpyModel *) PlayerSpyModel_copy(compat_type(PlayerSpyModel *) selfPtr) {
	stemobject_copy_implementation(PlayerSpyModel, initCopy)
}

void PlayerSpyModel_update(compat_type(PlayerSpyModel *) selfPtr) {
	PlayerSpyModel * self = selfPtr;
	
	PlayerModel_update(self);
	
	if (self->attackCooldown > 0) {
		self->attackCooldown--;
		self->facing = self->swordSwingFacing;
	}
	if (self->stealthCooldown > 0) {
		self->stealthCooldown--;
	}
}

bool PlayerSpyModel_atRest(compat_type(PlayerSpyModel *) selfPtr) {
	PlayerSpyModel * self = selfPtr;
	return self->attackCooldown == 0 &&
	       self->stealthCooldown == 0 &&
	       PlayerModel_atRest(self);
}

void PlayerSpyModel_startAttack(compat_type(PlayerSpyModel *) selfPtr) {
	PlayerSpyModel * self = selfPtr;
	
	if (self->attackCooldown == 0 && self->hitCooldown == 0) {
		self->swordSwingFacing = self->facing;
		switch (self->rank) {
			case 0:
			default:
				self->attackCooldown = 15;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, false, self->stealth ? 5 : 1, 0.0625f, 0.625f, 1.75f, 0.375f, 18));
				break;
			case 1:
				self->attackCooldown = 13;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, false, self->stealth ? 7 : 2, 0.078125f, 0.625f, 1.875f, 0.375f, 16));
				break;
			case 2:
				self->attackCooldown = 11;
				self->stateModel->addObject(self->stateModel, SwordSwingModel_create(self->stateModel, self, false, self->stealth ? 11 : 3, 0.09375f, 0.625f, 2.0f, 0.375f, 14));
				break;
		}
		if (self->stealth) {
			self->exitStealth(self);
		}
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_ATTACK_SMALL), NULL);
	}
}

void PlayerSpyModel_startSpecialAction(compat_type(PlayerSpyModel *) selfPtr) {
	PlayerSpyModel * self = selfPtr;
	
	if (self->stealth) {
		self->exitStealth(self);
	} else {
		self->enterStealth(self);
	}
}

void PlayerSpyModel_spotted(compat_type(PlayerSpyModel *) selfPtr, compat_type(EnemyModel *) enemy) {
	PlayerSpyModel * self = selfPtr;
	
	if (self->stealth) {
		self->exitStealth(self);
	}
	self->stealthCooldown = COMBAT_STEALTH_COOLDOWN;
}

void PlayerSpyModel_takeDamage(compat_type(PlayerSpyModel *) selfPtr, int amount, Direction fromDirection) {
	PlayerSpyModel * self = selfPtr;
	
	PlayerModel_takeDamage(self, amount, fromDirection);
	self->stealthCooldown = COMBAT_STEALTH_COOLDOWN;
	if (self->stealth) {
		self->exitStealth(self);
	}
}

void PlayerSpyModel_enterStealth(compat_type(PlayerSpyModel *) selfPtr) {
	PlayerSpyModel * self = selfPtr;
	
	if (self->stealthCooldown == 0 && self->attackCooldown == 0) {
		self->velocityMax = STEALTH_MAX_VELOCITIES[self->rank];
		self->stealth = true;
		self->eventDispatcher->dispatchEvent(self->eventDispatcher, Atom_fromString(PLAYER_EVENT_SPY_SNEAK), NULL);
	}
}

void PlayerSpyModel_exitStealth(compat_type(PlayerSpyModel *) selfPtr) {
	PlayerSpyModel * self = selfPtr;
	
	self->velocityMax = MAX_VELOCITIES[self->rank];
	self->stealth = false;
}
