/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/AudioManager.h"
#include "convergence/PlayerDuelistModel.h"
#include "convergence/PlayerModel.h"
#include "convergence/PlayerRangerModel.h"
#include "convergence/PlayerSpyModel.h"
#include "convergence/PlayerView.h"
#include <stddef.h>
#include <stdlib.h>

struct vertex_p2f {
	GLfloat position[2];
};

struct vertex_p2f_c3f {
	GLfloat position[2];
	GLfloat color[3];
};

PlayerView * PlayerView_create(compat_type(PlayerModel *) modelPtr) {
	stemobject_create_implementation(PlayerView, init, modelPtr);
}

static bool soundEffect(void * sender, Atom eventID, void * eventData, void * context) {
	AudioManager_play(context);
	return true;
}

static bool hit(void * sender, Atom eventID, void * eventData, void * context) {
	if (*(int *) eventData == 0) {
		AudioManager_play("player_block");
	} else {
		AudioManager_play("player_hit");
	}
	return true;
}

#define SHIELD_TESSELATIONS 12
#define SHIELD_MIN_RADIUS 0.125f
#define SHIELD_MAX_RADIUS 0.21875f

static char * swordSwingConst = "sword_swing";
static char * knifeConst = "knife";
static char * shootConst = "shoot";
static char * sneakConst = "sneak";
static char * rangerLockConst = "ranger_lock";
static char * dashConst = "dash";
static char * playerDeathConst = "player_death";
static char * playerHealedConst = "player_healed";
static char * gotKeyConst = "got_key";
static char * playerRankUpConst = "player_rank_up";

void PlayerView_init(compat_type(PlayerView *) selfPtr, compat_type(PlayerModel *) modelPtr) {
	PlayerView * self = selfPtr;
	PlayerModel * model = modelPtr;
	struct vertex_p2f_c3f healthVertices[8] = {
		{{-0.375f, -0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{-0.375f,  0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{ 0.0f,   -0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{ 0.0f,    0.03125f}, {1.0f, 1.0f, 1.0f}},
		{{ 0.0f,   -0.03125f}, {0.0f, 0.0f, 0.0f}},
		{{ 0.0f,    0.03125f}, {0.0f, 0.0f, 0.0f}},
		{{ 0.375f, -0.03125f}, {0.0f, 0.0f, 0.0f}},
		{{ 0.375f,  0.03125f}, {0.0f, 0.0f, 0.0f}}
	};
	GLushort healthIndexes[12] = {
		0, 2, 1,
		2, 3, 1,
		4, 6, 5,
		6, 7, 5
	};
	
	MovingEntityView_init(self, modelPtr);
	self->dispose = PlayerView_dispose;
	self->draw = PlayerView_draw;
	
	glGenBuffers(1, &self->healthVBOID);
	glGenBuffers(1, &self->healthEBOID);
	glBindBuffer(GL_ARRAY_BUFFER, self->healthVBOID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->healthEBOID);
	glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_c3f) * 8, healthVertices, GL_STREAM_DRAW);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * 12, healthIndexes, GL_STATIC_DRAW);
	
	if (model->playerClass == PLAYER_CLASS_DUELIST) {
		struct vertex_p2f duelistShieldVertices[(SHIELD_TESSELATIONS + 1) * 2];
		GLushort duelistShieldIndexes[SHIELD_TESSELATIONS * 6];
		size_t tesselationIndex;
		float angle;
		
		glGenBuffers(1, &self->duelistShieldVBOID);
		glGenBuffers(1, &self->duelistShieldEBOID);
		glBindBuffer(GL_ARRAY_BUFFER, self->duelistShieldVBOID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->duelistShieldEBOID);
		duelistShieldVertices[0].position[0] = (model->radius + SHIELD_MIN_RADIUS);
		duelistShieldVertices[0].position[1] = 0.0f;
		duelistShieldVertices[1].position[0] = (model->radius + SHIELD_MAX_RADIUS);
		duelistShieldVertices[1].position[1] = 0.0f;
		for (tesselationIndex = 0; tesselationIndex < SHIELD_TESSELATIONS; tesselationIndex++) {
			angle = M_PI * (tesselationIndex + 1) / SHIELD_TESSELATIONS;
			duelistShieldVertices[tesselationIndex * 2 + 2].position[0] = cos(angle) * (model->radius + SHIELD_MIN_RADIUS);
			duelistShieldVertices[tesselationIndex * 2 + 2].position[1] = sin(angle) * (model->radius + SHIELD_MIN_RADIUS);
			duelistShieldVertices[tesselationIndex * 2 + 3].position[0] = cos(angle) * (model->radius + SHIELD_MAX_RADIUS);
			duelistShieldVertices[tesselationIndex * 2 + 3].position[1] = sin(angle) * (model->radius + SHIELD_MAX_RADIUS);
			duelistShieldIndexes[tesselationIndex * 6 + 0] = tesselationIndex * 2 + 0;
			duelistShieldIndexes[tesselationIndex * 6 + 1] = tesselationIndex * 2 + 1;
			duelistShieldIndexes[tesselationIndex * 6 + 2] = tesselationIndex * 2 + 2;
			duelistShieldIndexes[tesselationIndex * 6 + 3] = tesselationIndex * 2 + 2;
			duelistShieldIndexes[tesselationIndex * 6 + 4] = tesselationIndex * 2 + 3;
			duelistShieldIndexes[tesselationIndex * 6 + 5] = tesselationIndex * 2 + 1;
		}
		glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f) * (SHIELD_TESSELATIONS + 1) * 2, duelistShieldVertices, GL_STATIC_DRAW);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * SHIELD_TESSELATIONS * 6, duelistShieldIndexes, GL_STATIC_DRAW);
		
	} else if (model->playerClass == PLAYER_CLASS_SPY) {
		struct vertex_p2f combatIndicatorVertices[8] = {
			{{-model->radius,            model->radius - 0.125f}},
			{{-model->radius + 0.125f,   model->radius}},
			{{ model->radius,            model->radius - 0.125f}},
			{{ model->radius - 0.125f,   model->radius}},
			{{-model->radius,           -model->radius + 0.125f}},
			{{-model->radius + 0.125f,  -model->radius}},
			{{ model->radius,           -model->radius + 0.125f}},
			{{ model->radius - 0.125f,  -model->radius}}
		};
		GLushort combatIndicatorIndexes[8] = {0, 1, 2, 3, 4, 5, 6, 7};
		
		glGenBuffers(1, &self->spyCombatVBOID);
		glGenBuffers(1, &self->spyCombatEBOID);
		glBindBuffer(GL_ARRAY_BUFFER, self->spyCombatVBOID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->spyCombatEBOID);
		glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f) * 8, combatIndicatorVertices, GL_STATIC_DRAW);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * 8, combatIndicatorIndexes, GL_STATIC_DRAW);
		
	} else if (model->playerClass == PLAYER_CLASS_RANGER) {
		struct vertex_p2f rangerLockVertices[3] = {
			{{ 0.0f,   model->radius + 0.15625f}},
			{{-0.125f, model->radius + 0.03125}},
			{{ 0.125f, model->radius + 0.03125}}
		};
		GLushort rangerLockIndexes[3] = {0, 1, 2};
		
		glGenBuffers(1, &self->rangerLockVBOID);
		glGenBuffers(1, &self->rangerLockEBOID);
		glBindBuffer(GL_ARRAY_BUFFER, self->rangerLockVBOID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->rangerLockEBOID);
		glBufferData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f) * 3, rangerLockVertices, GL_STATIC_DRAW);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(GLushort) * 3, rangerLockIndexes, GL_STATIC_DRAW);
	}
	
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_ATTACK_BIG), soundEffect, swordSwingConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_ATTACK_SMALL), soundEffect, knifeConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_SHOOT), soundEffect, shootConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_SPY_SNEAK), soundEffect, sneakConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANGER_LOCK), soundEffect, rangerLockConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DASH), soundEffect, dashConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_HIT), hit, self);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), soundEffect, playerDeathConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_HEALED), soundEffect, playerHealedConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_GOT_KEY), soundEffect, gotKeyConst);
	self->model->eventDispatcher->registerForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), soundEffect, playerRankUpConst);
}

void PlayerView_dispose(compat_type(PlayerView *) selfPtr) {
	PlayerView * self = selfPtr;
	
	glDeleteBuffers(1, &self->healthVBOID);
	glDeleteBuffers(1, &self->healthEBOID);
	if (((PlayerModel *) self->model)->playerClass == PLAYER_CLASS_DUELIST) {
		glDeleteBuffers(1, &self->duelistShieldVBOID);
		glDeleteBuffers(1, &self->duelistShieldEBOID);
		
	} else if (((PlayerModel *) self->model)->playerClass == PLAYER_CLASS_SPY) {
		glDeleteBuffers(1, &self->spyCombatVBOID);
		glDeleteBuffers(1, &self->spyCombatEBOID);
		
	} else if (((PlayerModel *) self->model)->playerClass == PLAYER_CLASS_RANGER) {
		glDeleteBuffers(1, &self->rangerLockVBOID);
		glDeleteBuffers(1, &self->rangerLockEBOID);
	}
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_ATTACK_BIG), soundEffect, swordSwingConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_ATTACK_SMALL), soundEffect, knifeConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_SHOOT), soundEffect, shootConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_SPY_SNEAK), soundEffect, sneakConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANGER_LOCK), soundEffect, rangerLockConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DASH), soundEffect, dashConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_HIT), hit, self);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_DEATH), soundEffect, playerDeathConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_HEALED), soundEffect, playerHealedConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_GOT_KEY), soundEffect, gotKeyConst);
	self->model->eventDispatcher->unregisterForEvent(self->model->eventDispatcher, Atom_fromString(PLAYER_EVENT_RANK_UP), soundEffect, playerRankUpConst);
	MovingEntityView_dispose(self);
}

void PlayerView_draw(compat_type(PlayerView *) selfPtr, Matrix modelviewMatrix, double deltaTime) {
	PlayerView * self = selfPtr;
	PlayerModel * model = (PlayerModel *) self->model;
	struct vertex_p2f_c3f vertices[4];
	float centerX;
	unsigned char alpha;
	
	if (model->playerClass == PLAYER_CLASS_DUELIST && ((PlayerDuelistModel *) model)->blocking && ((PlayerDuelistModel *) model)->attackCooldown == 0) {
		Matrix matrix;
		
		matrix = Matrix_rotated(Matrix_translated(modelviewMatrix, model->position.x, model->position.y, 0.0f), Vector3_init(0.0f, 0.0f, -1.0f), directionToRadians(model->facing));
		glBindBuffer(GL_ARRAY_BUFFER, self->duelistShieldVBOID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->duelistShieldEBOID);
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(matrix.m);
		glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f), (void *) offsetof(struct vertex_p2f, position));
		glEnableClientState(GL_VERTEX_ARRAY);
		glColor3ub(0xBF, 0xBF, 0xBF);
		glDrawElements(GL_TRIANGLES, SHIELD_TESSELATIONS * 6, GL_UNSIGNED_SHORT, 0);
		glDisableClientState(GL_VERTEX_ARRAY);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		
	} else if (model->playerClass == PLAYER_CLASS_SPY && ((PlayerSpyModel *) model)->stealthCooldown > 0) {
		Matrix matrix;
		
		matrix = Matrix_rotated(Matrix_translated(modelviewMatrix, model->position.x, model->position.y, 0.0f), Vector3_init(0.0f, 0.0f, -1.0f), directionToRadians(model->facing));
		glBindBuffer(GL_ARRAY_BUFFER, self->spyCombatVBOID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->spyCombatEBOID);
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(matrix.m);
		glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f), (void *) offsetof(struct vertex_p2f, position));
		glEnableClientState(GL_VERTEX_ARRAY);
		glColor3ub(0xFF, 0x00, 0x00);
		glDrawElements(GL_LINES, 8, GL_UNSIGNED_SHORT, 0);
		glDisableClientState(GL_VERTEX_ARRAY);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
		
	} else if (model->playerClass == PLAYER_CLASS_RANGER && ((PlayerRangerModel *) model)->facingLocked) {
		Matrix matrix;
		
		matrix = Matrix_rotated(Matrix_translated(modelviewMatrix, model->position.x, model->position.y, 0.0f), Vector3_init(0.0f, 0.0f, -1.0f), directionToRadians(model->facing));
		glBindBuffer(GL_ARRAY_BUFFER, self->rangerLockVBOID);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->rangerLockEBOID);
		glMatrixMode(GL_MODELVIEW);
		glLoadMatrixf(matrix.m);
		glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f), (void *) offsetof(struct vertex_p2f, position));
		glEnableClientState(GL_VERTEX_ARRAY);
		glColor3ub(0x00, 0xBF, 0x00);
		glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_SHORT, 0);
		glDisableClientState(GL_VERTEX_ARRAY);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
	}
	
	if (model->playerClass == PLAYER_CLASS_SPY && ((PlayerSpyModel *) model)->stealth) {
		alpha = 0x7F;
	} else {
		alpha = 0xFF;
	}
	if (model->hitCooldown > 0 && (model->playerClass != PLAYER_CLASS_DUELIST || ((PlayerDuelistModel *) model)->blockBuffRemaining == 0)) {
		self->outerColor = 0xFFFFFF << 8 | alpha;
		self->innerColor = 0xDFCF7F << 8 | alpha;
	} else {
		switch (model->playerClass) {
			case PLAYER_CLASS_BERSERKER:
				self->innerColor = 0xEF6F1F << 8 | alpha;
				self->outerColor = 0xC75D1A << 8 | alpha;
				break;
			case PLAYER_CLASS_RANGER:
				self->innerColor = 0x00FF00 << 8 | alpha;
				self->outerColor = 0x00BF00 << 8 | alpha;
				break;
			case PLAYER_CLASS_SPY:
				self->innerColor = 0xFFFF40 << 8 | alpha;
				self->outerColor = 0xBFBF30 << 8 | alpha;
				break;
			case PLAYER_CLASS_DUELIST:
				if (((PlayerDuelistModel *) model)->blockBuffRemaining > 0) {
					self->innerColor = 0xFF9FFF << 8 | alpha;
					self->outerColor = 0xB97FFF << 8 | alpha;
				} else {
					self->innerColor = 0xB97FFF << 8 | alpha;
					self->outerColor = 0xAF5FFF << 8 | alpha;
				}
				break;
			case PLAYER_CLASS_SPEEDSTER:
				self->innerColor = 0x1F2FFF << 8 | alpha;
				self->outerColor = 0x0000FF << 8 | alpha;
				break;
		}
	}
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	MovingEntityView_draw(self, modelviewMatrix, deltaTime);
	glDisable(GL_BLEND);
	
	centerX = -0.375f + 0.75f * model->health / model->maxHealth;
	vertices[0].position[0] = centerX;
	vertices[0].position[1] = -0.03125f;
	vertices[0].color[0] = 1.0f;
	vertices[0].color[1] = 1.0f;
	vertices[0].color[2] = 1.0f;
	vertices[1].position[0] = centerX;
	vertices[1].position[1] = 0.03125f;
	vertices[1].color[0] = 1.0f;
	vertices[1].color[1] = 1.0f;
	vertices[1].color[2] = 1.0f;
	vertices[2].position[0] = centerX;
	vertices[2].position[1] = -0.03125f;
	vertices[2].color[0] = 0.0f;
	vertices[2].color[1] = 0.0f;
	vertices[2].color[2] = 0.0f;
	vertices[3].position[0] = centerX;
	vertices[3].position[1] = 0.03125f;
	vertices[3].color[0] = 0.0f;
	vertices[3].color[1] = 0.0f;
	vertices[3].color[2] = 0.0f;
	
	glBindBuffer(GL_ARRAY_BUFFER, self->healthVBOID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self->healthEBOID);
	glBufferSubData(GL_ARRAY_BUFFER, sizeof(struct vertex_p2f_c3f) * 2, sizeof(struct vertex_p2f_c3f) * 4, vertices);
	
	Matrix_translate(&modelviewMatrix, model->position.x, model->position.y + 0.75f, 0.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	glVertexPointer(2, GL_FLOAT, sizeof(struct vertex_p2f_c3f), (void *) offsetof(struct vertex_p2f_c3f, position));
	glColorPointer(3, GL_FLOAT, sizeof(struct vertex_p2f_c3f), (void *) offsetof(struct vertex_p2f_c3f, color));
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glDrawElements(GL_TRIANGLES, 12, GL_UNSIGNED_SHORT, 0);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glBindBuffer(GL_ARRAY_BUFFER, 0);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
}
