/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "convergence/ResourceWiring.h"
#include "convergence/DungeonModel.h"
#include "convergence/DungeonRoomModel.h"
#include "convergence/DungeonTileColorMap.h"
#include "convergence/TextureCanvas.h"
#include "glbitmapfont/GLBitmapFont.h"
#include "gltexture/GLTexture.h"
#include "jsonserialization/JSONDeserializationContext.h"
#include "pngimageio/PNGImageIO.h"
#include "utilities/IOUtilities.h"
#include "vorbisaudioio/VorbisAudioIO.h"
#include <limits.h>
#include <stdio.h>

void * loadPNGFlip(const char * resourceName, void * context) {
	return PNGImageIO_loadPNGFile(resourceName, PNG_PIXEL_FORMAT_AUTOMATIC, true);
}

void * loadTexture(const char * resourceName, void * context) {
	ResourceManager * resourceManager = context;
	JSONDeserializationContext * deserializationContext;
	GLTexture * texture;
	BitmapImage * image;
	
	deserializationContext = JSONDeserializationContext_createWithFile(resourceName);
	if (deserializationContext->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Error: Couldn't load %s as a json file (error %d)\n", resourceName, deserializationContext->status);
		return NULL;
	}
	texture = GLTexture_deserialize(deserializationContext);
	if (texture == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a texture metadata file (error %d)\n", resourceName, deserializationContext->status);
		deserializationContext->dispose(deserializationContext);
		return NULL;
	}
	deserializationContext->dispose(deserializationContext);
	
	image = resourceManager->referenceResource(resourceManager, "png_flip", texture->imageName);
	if (image == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a png file\n", texture->imageName);
		texture->dispose(texture);
		return NULL;
	}
	texture->setImage(texture, 0, image->width, image->height, image->bytesPerRow, image->pixels);
	resourceManager->releaseResource(resourceManager, "png_flip", texture->imageName);
	
	return texture;
}

void * loadTextureCanvas(const char * resourceName, void * context) {
	JSONDeserializationContext * deserializationContext;
	TextureCanvas * textureCanvas;
	
	deserializationContext = JSONDeserializationContext_createWithFile(resourceName);
	if (deserializationContext->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Error: Couldn't load %s as a json file (error %d)\n", resourceName, deserializationContext->status);
		return NULL;
	}
	textureCanvas = TextureCanvas_deserialize(deserializationContext);
	if (textureCanvas == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a texture canvas file (error %d)\n", resourceName, deserializationContext->status);
		deserializationContext->dispose(deserializationContext);
		return NULL;
	}
	deserializationContext->dispose(deserializationContext);
	
	return textureCanvas;
}

void * loadBitmapFont(const char * resourceName, void * context) {
	ResourceManager * resourceManager = context;
	JSONDeserializationContext * deserializationContext;
	GLBitmapFont * font;
	GLTexture * texture;
	
	deserializationContext = JSONDeserializationContext_createWithFile(resourceName);
	if (deserializationContext->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Error: Couldn't load %s as a json file (error %d)\n", resourceName, deserializationContext->status);
		return NULL;
	}
	font = GLBitmapFont_deserialize(deserializationContext);
	if (font == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a font metadata file (error %d)\n", resourceName, deserializationContext->status);
		deserializationContext->dispose(deserializationContext);
		return NULL;
	}
	deserializationContext->dispose(deserializationContext);
	
	texture = resourceManager->referenceResource(resourceManager, "texture", font->textureName);
	if (texture == NULL) {
		font->dispose(font);
		return NULL;
	}
	font->setTexture(font, texture, false);
	
	return font;
}

void * loadDungeon(const char * resourceName, void * context) {
	JSONDeserializationContext * deserializationContext;
	DungeonModel * dungeon;
	
	deserializationContext = JSONDeserializationContext_createWithFile(resourceName);
	if (deserializationContext->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Error: Couldn't load %s as a json file (error %d)\n", resourceName, deserializationContext->status);
		return NULL;
	}
	dungeon = DungeonModel_deserialize(deserializationContext);
	if (dungeon == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a dungeon file (error %d)\n", resourceName, deserializationContext->status);
		deserializationContext->dispose(deserializationContext);
		return NULL;
	}
	deserializationContext->dispose(deserializationContext);
	
	return dungeon;
}

void * loadDungeonRoom(const char * resourceName, void * context) {
	ResourceManager * resourceManager = context;
	JSONDeserializationContext * deserializationContext;
	DungeonRoomModel * room;
	BitmapImage * image;
	DungeonTileColorMap * colorMap;
	
	deserializationContext = JSONDeserializationContext_createWithFile(resourceName);
	if (deserializationContext->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Error: Couldn't load %s as a json file (error %d)\n", resourceName, deserializationContext->status);
		return NULL;
	}
	room = DungeonRoomModel_deserialize(deserializationContext);
	if (room == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a dungeon room file (error %d)\n", resourceName, deserializationContext->status);
		deserializationContext->dispose(deserializationContext);
		return NULL;
	}
	deserializationContext->dispose(deserializationContext);
	
	deserializationContext = JSONDeserializationContext_createWithFile(room->colorMapFileName);
	if (deserializationContext->status != SERIALIZATION_ERROR_OK) {
		fprintf(stderr, "Error: Couldn't load %s as a json file (error %d)\n", room->colorMapFileName, deserializationContext->status);
		room->dispose(room);
		return NULL;
	}
	colorMap = DungeonTileColorMap_deserialize(deserializationContext);
	if (colorMap == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a dungeon tile color map file (error %d)\n", room->colorMapFileName, deserializationContext->status);
		deserializationContext->dispose(deserializationContext);
		room->dispose(room);
		return NULL;
	}
	deserializationContext->dispose(deserializationContext);
	
	image = resourceManager->referenceResource(resourceManager, "png_flip", room->tileMapFileName);
	if (image == NULL) {
		fprintf(stderr, "Error: Couldn't load %s as a png file\n", room->tileMapFileName);
		colorMap->dispose(colorMap);
		room->dispose(room);
		return NULL;
	}
	
	room->loadTileMap(room, colorMap, image->pixels, image->width, image->height, image->bytesPerRow, BitmapImage_pixelFormatBytes(image->pixelFormat));
	resourceManager->releaseResource(resourceManager, "png_flip", room->tileMapFileName);
	colorMap->dispose(colorMap);
	
	return room;
}

void * loadOggVorbis(const char * resourceName, void * context) {
	return VorbisAudioIO_loadOggVorbisFile(resourceName);
}

void unloadPNG(void * resource, void * context) {
	BitmapImage * image = resource;
	
	image->dispose(image);
}

void unloadTexture(void * resource, void * context) {
	GLTexture * texture = resource;
	
	texture->dispose(texture);
}

void unloadTextureCanvas(void * resource, void * context) {
	TextureCanvas * canvas = resource;
	
	canvas->dispose(canvas);
}

void unloadBitmapFont(void * resource, void * context) {
	GLBitmapFont * font = resource;
	ResourceManager * resourceManager = context;
	
	resourceManager->releaseResource(resourceManager, "texture", font->textureName);
	font->dispose(font);
}

void unloadDungeon(void * resource, void * context) {
	DungeonModel * dungeon = resource;
	
	dungeon->dispose(dungeon);
}

void unloadDungeonRoom(void * resource, void * context) {
	DungeonRoomModel * room = resource;
	
	room->dispose(room);
}

void unloadOggVorbis(void * resource, void * context) {
	PCMAudio * audio = resource;
	
	audio->dispose(audio);
}
