/*
  Copyright (c) 2011 Alex Diener
  
  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.
  
  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:
  
  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.
  
  Alex Diener adiener@sacredsoftware.net
*/

#include "3dmath/Matrix.h"
#include "convergence/SplashScreenState.h"
#include "convergence/SharedEvents.h"
#include "convergence/ShellStateGlobals.h"
#include "shell/Shell.h"
#include "shell/ShellKeyCodes.h"
#include <stdlib.h>
#include <string.h>

SplashScreenState * SplashScreenState_create(GameSession * gameSession) {
	stemobject_create_implementation(SplashScreenState, init, gameSession)
}

void SplashScreenState_init(compat_type(SplashScreenState *) selfPtr, GameSession * gameSession) {
	SplashScreenState * self = selfPtr;
	
	StemObject_init(self);
	self->dispose = SplashScreenState_dispose;
	self->activate = SplashScreenState_activate;
	self->deactivate = SplashScreenState_deactivate;
	self->draw = SplashScreenState_draw;
	
	self->gameSession = gameSession;
	self->selectedItemIndex = 0;
}

void SplashScreenState_dispose(compat_type(SplashScreenState *) selfPtr) {
	SplashScreenState * self = selfPtr;
	
	StemObject_dispose(self);
}

static void selectItem(SplashScreenState * self, int itemIndex) {
	switch (itemIndex) {
		case 0:
			self->stateController->transition(self->stateController, "choose_class");
			break;
			
		case 1:
			self->gameSession->showingInstructionsBeforeGame = false;
			self->stateController->transition(self->stateController, "instructions");
			break;
			
		case 2:
			self->stateController->transition(self->stateController, "preferences");
			break;
			
		case 3:
			exit(EXIT_SUCCESS);
			break;
	}
}

static bool mouseDown(void * sender, Atom eventID, void * eventData, void * context) {
	SplashScreenState * self = context;
	struct mouseEvent * event = eventData;
	float localY;
	
	if (event->button == 0) {
		localY = (g_windowHeight - event->y) / g_windowHeight * 2.0f - 1.0f;
		if (localY <= 0.375f) {
			if (localY > 0.125f) {
				selectItem(self, 0);
			} else if (localY > -0.125f) {
				selectItem(self, 1);
			} else if (localY > -0.375f) {
				selectItem(self, 2);
			} else {
				selectItem(self, 3);
			}
		}
	}
	return true;
}

static bool keyDown(void * sender, Atom eventID, void * eventData, void * context) {
	SplashScreenState * self = context;
	struct keyEvent * event = eventData;
	
	switch (event->keyCode) {
		case KEYBOARD_UP_ARROW:
			self->selectedItemIndex = (self->selectedItemIndex + 3) % 4;
			Shell_redisplay();
			return true;
			
		case KEYBOARD_DOWN_ARROW:
			self->selectedItemIndex = (self->selectedItemIndex + 1) % 4;
			Shell_redisplay();
			return true;
			
		case KEYBOARD_SPACEBAR:
		case KEYBOARD_RETURN_OR_ENTER:
			selectItem(self, self->selectedItemIndex);
			return true;
			
		case KEYBOARD_ESCAPE:
			exit(EXIT_SUCCESS);
			return true;
			
		default:
			return false;
	}
}

void SplashScreenState_activate(compat_type(SplashScreenState *) selfPtr) {
	SplashScreenState * self = selfPtr;
	
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->registerForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->font = self->gameSession->resourceManager->referenceResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
	Shell_redisplay();
}

void SplashScreenState_deactivate(compat_type(SplashScreenState *) selfPtr) {
	SplashScreenState * self = selfPtr;
	
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_MOUSE_DOWN), mouseDown, self);
	self->stateController->eventDispatcher->unregisterForEvent(self->stateController->eventDispatcher, Atom_fromString(EVENT_KEY_DOWN), keyDown, self);
	self->gameSession->resourceManager->releaseResource(self->gameSession->resourceManager, "bitmapfont", "font.json");
}

static void drawString(SplashScreenState * self, bool highlighted, const char * string, float y, float textHeight) {
	float stringWidth;
	
	if (highlighted) {
		glColor4f(1.0f, 0.5f, 0.0f, 1.0f);
	} else {
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	}
	stringWidth = self->font->measureString(self->font, string, strlen(string));
	self->font->drawString(self->font, string, strlen(string), textHeight, -stringWidth * textHeight * 0.5f, y - textHeight * 0.5f, 0.0f);
}

void SplashScreenState_draw(compat_type(SplashScreenState *) selfPtr) {
	SplashScreenState * self = selfPtr;
	float ratio;
	Matrix modelviewMatrix, projectionMatrix;
	
	glClear(GL_COLOR_BUFFER_BIT);
	ratio = (float) g_windowWidth / g_windowHeight;
	projectionMatrix = Matrix_ortho(Matrix_identity(), -ratio, ratio, -1.0f, 1.0f, -1.0f, 1.0f);
	modelviewMatrix = Matrix_identity();
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projectionMatrix.m);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix.m);
	
	drawString(self, false, "CONVERGENCE", 0.75f, 0.25f);
	drawString(self, self->selectedItemIndex == 0, "Start game", 0.25f, 0.125f);
	drawString(self, self->selectedItemIndex == 1, "Instructions", 0.0f, 0.125f);
	drawString(self, self->selectedItemIndex == 2, "Preferences", -0.25f, 0.125f);
	drawString(self, self->selectedItemIndex == 3, "Quit", -0.75f, 0.125f);
}
